#!/usr/bin/env python3
import csv
from pprint import pprint
from pathlib import Path
import rdflib
import ontquery as oq
from pyontutils.utils import byCol, relative_path, noneMembers
from pyontutils.core import resSource
from pyontutils.namespaces import OntCuries
from pyontutils.namespaces import interlex_namespace, definition, NIFRID, NIFSTD
from pyontutils.namespaces import rdfs
# import these last so that graphBase resets (sigh)
from neurondm.core import auth
from neurondm.lang import *
from neurondm import *
from neurondm.phenotype_namespaces import BBP, CUT, Layers, Regions

try:
    breakpoint
except NameError:
    # <=python-3.7
    from IPython import embed as breakpoint

# TODO
# 1. inheritance for owlClass from python classes
# 2. add ttl serialization for subclasses of EBM
# 3. pv superclass for query example

class PT(OntTerm):  # FIXME subclassing very broken :/ probably because __real_init__ is a __ method
    def __repr__(self):
        if self.label:
            return f'Phenotype({self.curie}, ilxtr.hasSomaLocatedIn)  # {self.label}'
        else:
            return super().__repr__()


OntTerm.set_repr_args('curie', 'label')


class NeuronSWAN(NeuronEBM):
    owlClass = OntId('ilxtr:NeuronSWAN').u


rename_rules = {'Colliculus inferior': 'Inferior colliculus',
                'Colliculus superior': 'Superior colliculus',
                'Premammillary nucleus dorsal': 'Dorsal premammillary nucleus',
                'Premammillary nucleus ventral': 'Ventral premammillary nucleus',
                'Septal complex lateral': 'Lateral septal complex',
                'Septal complex medial': 'Medial septal complex',
                'Substantia nigra pars reticulata': 'Reticular part of substantia nigra',
                'Thalamic reticular nucleus': 'Reticular thalamic nucleus',
                'Trigeminal nerve motor nucleus': 'Motor nucleus of trigeminal nerve',
                'Trigeminal nerve principal sensory nucleus': 'Principal sensory nucleus of trigeminal nerve',
                'Dorsal root ganglion cell': 'Dorsal root ganglion A alpha-beta non-nociceptive neuron',
                'Neocortex layer 2-3 pyramidal cell': 'Neocortex pyramidal layer 2-3 cell',
                #'Neocortex layer 5 pyramidal cell':  # TODO layer 5-6??
                'Hippocampus CA2 Basket cell': 'Hippocampus CA2 basket cell broad',
                'Neocortex layer 4 spiny stellate cell': 'Neocortex stellate layer 4 cell',

                # some of the missing neurolex neurons, this mapping doesn't quite work, but is correct
                'Neostriatum Neurogliaform cell': 'Striatum neurogliaform cell',
                'Neostriatum SOM/NOS cell': 'Striatum somatostatin positive nitric oxide synthase cell',
                'Neostriatum cholinergic cell': 'Striatum cholinergic cell',
                'Neostriatum gaba/parvalbumin interneuron': 'Striatum parvalbumin interneuron',
                #Neostriatum direct pathway spiny neuron: Striatum medium spiny neuron,
                #Neostriatum indirect pathway spiny neuron: Striatum medium spiny neuron,
}


def make_contains_rules():
    contains_rules = dict(GABAergic=CUT.GABA,
                          cholinergic=CUT.ACh,
                          glutamatergic=CUT.Glu,
                          serotonergic=CUT.Ser,
                          Cholinergic=CUT.ACh,
                          Glutamatergic=CUT.Glu,
                          Serotonergic=CUT.Ser,
                          #principle=CUT.proj,  # NOTE this was a spelling error
                          projection=CUT.proj,
                          intrinsic=CUT.intrinsic,
                          interneuron=CUT.intrinsic,
                          Striatum=Phenotype('UBERON:0002435', ilxtr.hasSomaLocatedIn),
                          #CA1=Regions.CA1,
                          #CA2=Regions.CA2,
                          #CA3=Regions.CA3,
                          trilaminar=BBP.TRI,
                          neurogliaform=BBP.NGC,  # FIXME
                          parvalbumin=CUT.PV,  # FIXME
                          calbindin=CUT.CB,
                          calretinin=CUT.CR,
                          cholecystokinin=CUT.CCK,
                          somatostatin=CUT.SST,
                          orexin=Phenotype('PR:000008476', ilxtr.hasExpressionPhenotype),
                          oxytocin=Phenotype('CHEBI:7872', ilxtr.hasExpressionPhenotype),
                          bitufted=BBP.BTC,
                          radiatum=Layers.SR,

                          #motor=Phenotype(ilxtr.MotorPhenotype, ilxtr.hasCircuitRolePhenotype),
    )

    contains_rules.update({  # FIXME still need to get some of the original classes from neurolex
        'medium spiny': (BBP.MSN, Phenotype(ilxtr.SpinyPhenotype, ilxtr.hasDendriteMorphologicalPhenotype)),
        'TH+': CUT.TH,
        'Thalamic reticular nucleus': CUT.TRN,  # FIXME disambiguate with reticular nucleus
        'Midbrain reticular nucleus': CUT.MRN,
        'Ambiguous nucleus': Phenotype('UBERON:0001719', ilxtr.hasSomaLocatedIn),
        'Accumbens nucleus': Phenotype('UBERON:0001882', ilxtr.hasSomaLocatedIn),
        'Neocortex layer I': Layers.L1, # FIXME consistency in naming?
        'Neocortex layer 2': Layers.L2,
        'Neocortex layer 3': Layers.L3,
        'Neocortex layer 4': Layers.L4,
        'Neocortex layer 5': Layers.L5,
        'Neocortex layer 6': Layers.L6,
        'cortex layer III': Layers.L3,  # probably needs the cortex to still be attached
        'cortex layer II': Layers.L2,  # FIXME medial entorhinal ...
        'neuropeptide Y': CUT.NPY,
        'vasoactive intestinal peptide': CUT.VIP,
        'Pedunculopontine nucleus': Phenotype('UBERON:0002142', ilxtr.hasSomaLocatedIn),
        #'Incertus nucleus': OntTerm,
        'Raphe nucleus medial': OntTerm,
        'double bouquet': BBP.DBC,
        'Neocortex  ': Regions.CTX,
        'Cuneate nucleus  ': Phenotype('UBERON:0002045', ilxtr.hasSomaLocatedIn),
        'Interstitial nucleus of Cajal  ': Phenotype('UBERON:0002551', ilxtr.hasSomaLocatedIn),
        'Suprachiasmatic nucleus  ': Phenotype('UBERON:0002034', ilxtr.hasSomaLocatedIn),
        'Darkshevich nucleus  ': Phenotype('UBERON:0002711', ilxtr.hasSomaLocatedIn),
        'Spinal cord ventral horn  ': Phenotype('UBERON:0002257', ilxtr.hasSomaLocatedIn),  # FIXME ??? II just hanging out!?
        'Area postrema  ': Phenotype('UBERON:0002162', ilxtr.hasSomaLocatedIn),
        'Sublingual nucleus  ': Phenotype('UBERON:0002881', ilxtr.hasSomaLocatedIn),
        'Arcuate nucleus medulla  ': Phenotype('UBERON:0002865', ilxtr.hasSomaLocatedIn),
        'Hippocampus CA1': Regions.CA1,
        'Hippocampus CA2': Regions.CA2,
        'Hippocampus CA3': Regions.CA3,
        'Hippocampus  ': Phenotype('UBERON:0001954', ilxtr.hasSomaLocatedIn),  # FIXME clarity here
        'Retrotrapezoid nucleus  ': Phenotype('UBERON:0009918', ilxtr.hasSomaLocatedIn),
        'Salivatory nucleus  ': Phenotype('UBERON:0004133', ilxtr.hasSomaLocatedIn),
        'Gracile nucleus  ': Phenotype('UBERON:0002161', ilxtr.hasSomaLocatedIn),
        'Laterodorsal tegmental nucleus  ': Phenotype('UBERON:0002267', ilxtr.hasSomaLocatedIn),
        'Medial entorhinal  ': Phenotype('UBERON:0007224', ilxtr.hasSomaLocatedIn),
        'Vagus dorsal motor nucleus': Phenotype('UBERON:0002870', ilxtr.hasSomaLocatedIn),
        'Bed nuclei of terminal stria': Phenotype('UBERON:0001880', ilxtr.hasSomaLocatedIn, label='bed nucleus of stria terminalis'),
        'Olfactory cortex': Phenotype('UBERON:0002894', ilxtr.hasSomaLocatedIn, label='olfactory cortex'),
        'Globus pallidus external segment': Phenotype('UBERON:0002476', ilxtr.hasSomaLocatedIn,  label='lateral globus pallidus'),
        'Globus pallidus internal segment': Phenotype('UBERON:0002477', ilxtr.hasSomaLocatedIn,  label='medial globus pallidus'),
        'Septal complex lateral': Phenotype('UBERON:0007628', ilxtr.hasSomaLocatedIn, label='lateral septal complex'),  # FIXME missing synonym in uberon
        'Septal complex medial': Phenotype('UBERON:0007629', ilxtr.hasSomaLocatedIn, label='medial septal complex'),  # FIXME missing synonym in uberon
        'Premammillary nucleus dorsal': OntTerm('UBERON:0007767', label='dorsal premammillary nucleus').asPhenotype(),  # FIXME missing synonym in uberon
        'Premammillary nucleus ventral': OntTerm('UBERON:0007768', label='ventral premammillary nucleus').asPhenotype(),  # FIXME missing synonym in uberon
        'Raphe nucleus dorsal': OntTerm('UBERON:0002043', label='dorsal raphe nucleus').asPhenotype(),  # FIXME missing synonym in uberon
        'Raphe nucleus medial': OntTerm('UBERON:0003004', label='median raphe nucleus').asPhenotype(),  # FIXME missing synonym in uberon
        'Periventricular hypothalamic zone': OntTerm('UBERON:0002271', label='periventricular zone of hypothalamus').asPhenotype(),  # FIXME missing synonym in uberon
        'Lateral lemniscus nuclear complex': OntTerm('UBERON:0006840', label='nucleus of lateral lemniscus').asPhenotype(),  # FIXME wow this was hard to find :/ may synonyms needed
        # 'nuclear complex of lateral lemniscus'
        # 'lateral lemniscus nucleus'
        'Medial hypothalamic zone': OntTerm('UBERON:0002272', label='medial zone of hypothalamus').asPhenotype(),  # SciGraph default search is really bad >_<
        'Incertus nucleus': OntTerm('UBERON:0035973', label='nucleus incertus').asPhenotype(),  # WOW ... that was bad
        'Retroambiguous nucleus': OntTerm('UBERON:0016848', label='retroambiguus nucleus').asPhenotype(),  # spelling ...
        'Trigeminal nerve motor nucleus': OntTerm('UBERON:0002633', label='motor nucleus of trigeminal nerve').asPhenotype(),  # had to cook up my search function to find this
        'Trigeminal nerve principal sensory nucleus': OntTerm('UBERON:0002597', label='principal sensory nucleus of trigeminal nerve').asPhenotype(),  # had to go to wikipedia for this one, and no amount of tweaking seems to help scigraph (sigh)
        'Pontine nuclei': OntTerm('UBERON:0002151', label='pontine nuclear group').asPhenotype(),  # pontine nuclei has a search collision in scigraph with UBERON:0002597 likely due to inclusion of overly broad synonyms on UBERON:0002597
        'Basolateral amygdalar complex': OntTerm('UBERON:0006107', label='basolateral amygdaloid nuclear complex').asPhenotype(),  # oof
        'Amygdala lateral': OntTerm('UBERON:0002886', label='lateral amygdaloid nucleus').asPhenotype(),  # conclusion: scigraph cannot handle reordering
        'Globus pallidus ventral': OntTerm('UBERON:0002778', label='ventral pallidum').asPhenotype(),
        'Habenula nuclei': OntTerm('UBERON:0008993', label='habenular nucleus').asPhenotype(),  # scigraph can't unpluralize nucleus?
        'Trapezoid Body medial nucleus': OntTerm('UBERON:0002833', label='medial nucleus of trapezoid body').asPhenotype(),
        'Gigantocellular reticular nucleus': OntTerm('UBERON:0002155', label='gigantocellular nucleus').asPhenotype(),
        'Bed nucleus of the stria terminalis': OntTerm('UBERON:0001880', label='bed nucleus of stria terminalis').asPhenotype(),
        'Hypothalamus paraventricular nucleus': OntTerm('UBERON:0001930', label='paraventricular nucleus of hypothalamus').asPhenotype(),
        'Hypothalamus tuberomammillary nucleus': OntTerm('UBERON:0001936', label='tuberomammillary nucleus').asPhenotype(),
        'Subthalamic nucleus': OntTerm('UBERON:0001906', label='subthalamic nucleus').asPhenotype(),

        'Spinocerebellar dorsal tract': OntTerm('UBERON:0002753', label='posterior spinocerebellar tract').asPhenotype(),
        'Spinocerebellar ventral tract': OntTerm('UBERON:0002987', label='anterior spinocerebellar tract').asPhenotype(),
        'Abducens nucleus': OntTerm('UBERON:0002682', label='abducens nucleus').asPhenotype(),
        'Medial amygdalar nucleus': OntTerm('UBERON:0002892', label='medial amygdaloid nucleus').asPhenotype(),  # a creeping madness issue
        'Hippocampal formation': OntTerm('UBERON:0002421', label='hippocampal formation').asPhenotype(),

    })

    for k, v in contains_rules.items():  # ah lack of types
        if v == OntTerm:
            continue

        if isinstance(v, tuple):
            for _v in v:
                if not isinstance(_v, graphBase):
                    raise TypeError(f'{k!r}: {_v!r}  # is not a Phenotype or some such')

        elif not isinstance(v, graphBase):
            raise TypeError(f'{k!r}: {v!r}  # is not a Phenotype or some such')

    return contains_rules


exact_rules = {'pyramidal cell': BBP.PC,
               'Neocortex neuron': Regions.CTX,
               'Neocortex': Regions.CTX,
               'Thalamic': CUT.Thal,
               'principal cell': CUT.proj,
               'principal neuron': CUT.proj,
               'Hypothalamus': Phenotype('UBERON:0001898', ilxtr.hasSomaLocatedIn, label='hypothalamus'),
}
terminals = 'cell', 'Cell', 'neuron', 'neurons', 'positive cell'  # TODO flag cell and neurons for inconsistency

def skip_pred(p):
    """ probably don't need to use this """
    return False
    #if 'ConnectionDetermined' in p:
        #return True


def fixname(name):
    return (name.
            replace('/', '-').
            replace(' ', '-').
            replace('(','-').
            replace(')', '-').
            replace('+','-'))


def make_cut_id(label):
    return 'TEMP:' + fixname(label)


def zap(pes):
    for pe in pes:
        # NOTE: these are RAW phenos that have not been converted
        if pe not in (Phenotype('BIRNLEX:212', ilxtr.hasTaxonRank),
                      Phenotype('NCBITaxon:7742', ilxtr.hasTaxonRank),
                      Phenotype('BIRNLEX:252', ilxtr.hasTaxonRank),
                      Phenotype('BIRNLEX:263', ilxtr.hasTaxonRank),
                      Phenotype('BIRNLEX:516', ilxtr.hasTaxonRank),):
            yield pe


def export_for_review(config, unmapped, partial, nlx_missing,
                      filename='cuts-review.csv',
                      with_curies=False):
    neurons = sorted(config.neurons())
    predicates = sorted(set(e for n in neurons
                            for me in n.edges
                            for e in (me if isinstance(me, tuple) else (me,))))  # columns
    q = graphBase.core_graph.transitive_subjects(rdfs.subPropertyOf, ilxtr.hasPhenotype)
    all_predicates = set(s for s in q)
    extra_predicates = sorted(p for p in all_predicates if p not in predicates)

    col_labels = {p.e:p.eLabel for n in neurons
                  for mp in n.pes
                  for p in (mp.pes if isinstance(mp, LogicalPhenotype) else (mp,))}

    header = (['curie', 'label'] +
              [col_labels[p] for p in predicates] +
              ['Status', 'PMID', 'synonyms', 'definition'] +
              [OntId(p).suffix for p in extra_predicates if not skip_pred(p)])

    def neuron_to_review_row(neuron, cols=predicates):  # TODO column names
        _curie = neuron.ng.qname(neuron.id_)
        curie = None if 'TEMP:' in _curie else _curie
        row = [curie, neuron.origLabel]
        for pdim in cols:  # pdim -> phenotypic dimension
            if pdim in neuron:
                #print('>>>>>>>>>>>>>', pdim, neuron)
                #if any(isinstance(p, LogicalPhenotype) for p in neuron):
                    #breakpoint()
                row.append(','.join(sorted([f'{_._pClass.qname}|{_.pLabel}'
                                            if with_curies else
                                            _.pLabel
                                            for _ in neuron[pdim]] if
                                           isinstance(neuron[pdim], list) else
                                           [neuron[pdim].pLabel])))
                #if col == ilxtr.hasLayerLocationPhenotype:
                    #derp = neuron[col]
                    #log = [p for p in derp if isinstance(p, LogicalPhenotype)]
                    #if log:
                        #print(log, row)
                        #breakpoint()
            else:
                row.append(None)

        return row

    #[n for n in neurons]
    resources = auth.get_path('resources')
    reviewcsv = resources / filename
    rows = [neuron_to_review_row(neuron) for neuron in neurons]

    for i, row in enumerate(rows):
        label = row[1]
        if label in unmapped:
            row.append('Unmapped')
        elif label in partial:
            rem = partial[label]
            row.append(f'Partial: {rem!r}')

        if label in nlx_missing:
            row.append('Could not find NeuroLex mapping')
        else:
            row.append(None)

        row.append(None)  # pmid
        if i < len(neurons):
            n = neurons[i]
            # FIXME
            row.append(','.join(n.config.out_graph[n.id_:NIFRID.synonym:]))  # syn
            row.append(','.join(n.config.out_graph[n.id_:definition:]))  # def

    rows = sorted(rows, key=lambda r:(1 if r[1] is None else 0, str(r[1])))
    incomplete = [[None, u] + [None] * (len(rows[0]) - 2) + ['Unmapped', None, None] for u in unmapped]
    incomplete = sorted(incomplete, key=lambda r:r[1])
    rows += incomplete
    with open(reviewcsv.as_posix(), 'wt') as f:
        writer = csv.writer(f, lineterminator='\n')
        writer.writerow(header)
        writer.writerows(rows)

    return [header] + rows


def get_smatch(labels_set2):
    contains_rules = make_contains_rules()
    skip = set()
    smatch = set()
    rem = {}
    for l in labels_set2:
        pes = tuple()
        l_rem = l
        for match, pheno in sorted(contains_rules.items(), key=lambda ab:-len(ab[0])):
            if not l_rem:
                break

            if len(match) > len(l_rem):
                continue

            t = None
            if match not in skip and pheno == OntTerm:
                try:
                    t = OntTerm(term=match)
                    log.debug(f'WTF {match} {t}')
                    if t.validated:
                        pheno = Phenotype(t.u, ilxtr.hasSomaLocatedIn)
                    else:
                        pheno = None
                except oq.exceptions.NotFoundError:
                    skip.add(match)
                    pheno = None

            if match in skip and pheno == OntTerm:
                pheno = None

            if match in l_rem and pheno:
                l_rem = l_rem.replace(match, '').strip()
                pes += (pheno if isinstance(pheno, tuple) else (pheno,))

        if l_rem in exact_rules:
            pes += (exact_rules[l_rem],)
            l_rem = ''

        if l_rem == '  neuron':
            l_rem = ''
        elif l_rem.endswith('  cell'):
            l_rem = l_rem[:-len('  cell')]
            #print('l_rem no cell:', l_rem)
        elif l_rem.endswith('  neuron'):
            l_rem = l_rem[:-len('  neuron')]
            #print('l_rem no neuron:', l_rem)

        hrm = [pe for pe in pes if pe.e == ilxtr.hasSomaLocatedIn]
        if '  ' in l_rem:
            #print('l_rem:', l_rem)
            #breakpoint()
            maybe_region, rest = l_rem.split('  ', 1)
        elif noneMembers(l_rem, *terminals) and not hrm:
            maybe_region, rest = l_rem, ''
            #print('MR:', maybe_region)
        else:
            #print(hrm)
            maybe_region = None

        if maybe_region:
            prefix_rank = ('UBERON', 'SWAN', 'BIRNLEX', 'SAO', 'NLXANAT', 'NLX')
            def key(ot):
                ranked = ot.prefix in prefix_rank
                qargs = ot._query_result._QueryResult__query_args
                if 'term' in qargs and qargs['term'] is not None:
                    arg = qargs['term'].lower()
                else:
                    arg = None
                return (not ranked,
                        prefix_rank.index(ot.prefix) if ranked else 0,
                        not (arg == ot.label.lower()))

            #ots = sorted((term for term in OntTerm.query(term=maybe_region,
                                                         #exclude_prefix=('FMA', 'NLX'))), key=key)

            #if not ots:
            ots = sorted((term for term in OntTerm.query(term=maybe_region,
                                                         exclude_prefix=('FMA',))), key=key)
            if not ots:
                log.error(f'No match for {maybe_region!r}')
            else:
                t = ots[0]
                if 'oboInOwl:id' in t.predicates:  # uberon replacement
                    t = OntTerm(t.predicates['oboInOwl:id'])

                t.set_next_repr('curie', 'label')
                log.info(f'Match for {maybe_region!r} was {t!r}')
                if t.validated:
                    l_rem = rest
                    pheno = Phenotype(t.u, ilxtr.hasSomaLocatedIn)  # FIXME
                    pes += (pheno,)

        if pes:
            smatch.add(l)

            if not l_rem or l_rem in ('neuron', 'neurons', 'cell', 'Cell', 'positive cell'):
                with NeuronCUT(CUT.Mammalia):
                    NeuronCUT(*zap(pes), id_=make_cut_id(l), label=l, override=True)
            else:
                rem[l] = l_rem

    return smatch, rem


def ontneurons(remote):
    in_config = Config(imports=[remote.iri + 'ttl/NIF-Cell.ttl',
                                remote.iri + 'ttl/NIF-Neuron-BrainRegion-Bridge.ttl',
                                remote.iri + 'ttl/NIF-Neuron-NT-Bridge.ttl'])

    eids = ['NLXNEURNT:090803',
            'NLXNEURNT:090804',
            'NLXNEURNT:090802',
            'NLXNEURNT:090807',
            'NLXCELL:0912004',
            # 'NLXCELL:0912003',  # leaving out since mismatch in modelling and definition
    ]

    nrns = [NeuronCUT(id_=eid) for eid in eids]
    inrns =  [n.asIndicator() for n in nrns]
    ont_config = Config('ontology-neurons')
    return ont_config


def main():
    branch=auth.get('neurons-branch')
    remote = OntId('NIFTTL:') if branch == 'master' else OntId(f'NIFRAW:{branch}/')

    ont_config = ontneurons(remote)
    ont_neurons = ont_config.neurons()

    bn_config = Config('basic-neurons',
                       # FIXME this should probably be pulled in automatically
                       # from the import statements, and it doesn't work even as is
                       # also a chicken and an egg problem here
                       imports=[remote.iri + 'ttl/generated/swanson.ttl'])

    #RDFL = oq.plugin.get('rdflib')  # FIXME ick
    #rdfl = RDFL(bn_config.core_graph, OntId)
    #OntTerm.query.ladd(rdfl)  # FIXME ick
    bn_config.load_existing()
    bn_neurons = bn_config.neurons()
    #OntTerm.query._services = OntTerm.query._services[:-1]  # FIXME ick

    ndl_config = Config('neuron_data_lifted')
    ndl_config.load_existing()  # FIXME this is extremely slow
    ndl_neurons = sorted(ndl_config.neurons())

    resources = auth.get_path('resources')
    cutcsv = resources / 'cut-development.csv'
    with open(cutcsv.as_posix(), 'rt') as f:
        rows = [l for l in csv.reader(f)]

    bc = byCol(rows)

    (_, *labels), *_ = zip(*bc)
    labels_set0 = set(labels)
    ns = []
    skipped = []
    bamscok = (NIFSTD.BAMSC1125,)
    for n in (ont_neurons + ndl_neurons):
        if n.id_ and 'BAMSC' in n.id_:
            if n.id_ not in bamscok:
                skipped.append(n)
                continue

        l = str(n.origLabel)
        if l is not None:
            for replace, match in rename_rules.items():  # HEH
                l = l.replace(match, replace)

        if l in labels:
            n._origLabel = l
            ns.append(n)

    ns = sorted(ns)
    sns = set(n.origLabel for n in ns)

    labels_set1 = labels_set0 - sns

    agen = [c.label for c in bc if c.autogenerated]
    sagen = set(agen)
    added = [c.label for c in bc if c.added]
    sadded = set(added)
    ans = []
    sans = set()
    missed = set()
    _bl = []  # XXX NOTE THE CONTINUE BELOW
    for n in bn_neurons:
        continue  # we actually get all of these with uberon, will map between them later
        # can't use capitalize here because there are proper names that stay uppercase
        l = n.label.replace('(swannt) ',
                            '').replace('Intrinsic',
                                        'intrinsic').replace('Projection',
                                                             'projection')

        for replace, match in rename_rules.items():  # HEH
            l = l.replace(match, replace)

        if l in agen:
            n._origLabel = l
            ans.append(n)
            sans.add(l)

        else:
            missed.add(l)

        _bl.append(l)

    agen_missing = sagen - sans
    labels_set2 = labels_set1 - sans

    nlx_labels = [c.label for c in bc if c.neurolex]
    snlx_labels = set(nlx_labels)

    class SourceCUT(resSource):
        sourceFile = 'nifstd/resources/cut-development.csv'  # FIXME relative to git workingdir...
        source_original = True

    sources = SourceCUT(),
    swanr = rdflib.Namespace(interlex_namespace('swanson/uris/readable/'))
    SWAN = interlex_namespace('swanson/uris/neuroanatomical-terminology/terms/')
    SWAA = interlex_namespace('swanson/uris/neuroanatomical-terminology/appendix/')
    config = Config('cut-development-raw', sources=sources, source_file=relative_path(__file__),
                    prefixes={'swanr': swanr,
                              'SWAN': SWAN,
                              'SWAA': SWAA,})
    ins = [None if OntId(n.id_).prefix == 'TEMP' else n.id_ for n in ns]
    ians = [None] * len(ans)

    with NeuronCUT(CUT.Mammalia):
        mamns = [NeuronCUT(*zap(n.pes), id_=i, label=n._origLabel, override=bool(i)).adopt_meta(n)
                 for i, n in zip(ins + ians, ns + ans)]

    smatch, rem = get_smatch(labels_set2)

    labels_set3 = labels_set2 - smatch
    added_unmapped = sadded & labels_set3

    # TODO preserve the names from neuronlex on import ...
    Neuron.write()
    Neuron.write_python()
    raw_neurons = config.neurons()
    # do this before creating the new config
    # even though we are in theory tripling number of neurons in the current config graph
    # it won't show up in the next config (and this is why we need to reengineer)
    raw_neurons_ind_undep = [n.asUndeprecated().asIndicator() for n in raw_neurons]
    config = Config('cut-development', sources=sources, source_file=relative_path(__file__),
                    prefixes={'swanr': swanr,
                              'SWAN': SWAN,
                              'SWAA': SWAA,})
    # FIXME the call to asUndprecated currenlty triggers addition
    # to the current config and output graph as a side effect (ick!)
    ids_updated_neurons = [n.asUndeprecated() for n in raw_neurons]
    assert len(ids_updated_neurons) == len(raw_neurons)
    Neuron.write()
    Neuron.write_python()
    progress = (len(labels_set0), len(sns), len(sans), len(smatch),
                len(labels_set1), len(labels_set2), len(labels_set3))
    prog_report = ('\nProgress:\n'
                   f'total:            {progress[0]}\n'
                   f'from nlx:         {progress[1]}\n'
                   f'from basic:       {progress[2]}\n'
                   f'from match:       {progress[3]}\n'
                   f'TODO after nlx:   {progress[4]}\n'
                   f'TODO after basic: {progress[5]}\n'
                   f'TODO after match: {progress[6]}\n')
    print(prog_report)
    assert progress[0] == progress[1] + progress[4], 'neurolex does not add up'
    assert progress[4] == progress[2] + progress[5], 'basic does not add up'

    lnlx = set(n.lower() for n in snlx_labels)
    sos = set(n.origLabel.lower() if n.origLabel else None for n in ndl_neurons)  # FIXME load origLabel
    nlx_review = lnlx - sos
    nlx_missing = sorted(nlx_review)
    print(f'\nNeuroLex listed as source but no mapping (n = {len(nlx_review)}):')
    _ = [print(l) for l in nlx_missing]

    partial = {k:v for k, v in rem.items() if v and v not in terminals}
    print(f'\nPartially mapped (n = {len(partial)}):')
    if partial:
        mk = max((len(k) for k in partial.keys())) + 2
        for k, v in sorted(partial.items()):
            print(f'{k:<{mk}} {v!r}')
            #print(f'{k!r:<{mk}}{v!r}')
        #pprint(partial, width=200)
    unmapped = sorted(labels_set3)
    print(f'\nUnmapped (n = {len(labels_set3)}):')
    _ = [print(l) for l in unmapped]

    no_location = [n for n in Neuron.neurons()
                   if noneMembers((ilxtr.hasSomaLocatedIn, ilxtr.hasSomaLocatedInLayer), *n.unique_predicates)]
    if __name__ == '__main__':
        review_rows = export_for_review(config, unmapped, partial, nlx_missing)
        breakpoint()

    return config, unmapped, partial, nlx_missing


if __name__ == '__main__':
    main()



from pprint import pprint
from contextlib import nullcontext  # 3.7
from functools import wraps
import rdflib  # FIXME decouple
import ontquery as oq
from hyputils.hypothesis import idFromShareLink, shareLinkFromId
from pyontutils import sheets
from pyontutils.sheets import update_sheet_values, get_note, Sheet
from pyontutils.scigraph import Vocabulary
from pyontutils.namespaces import ilxtr, TEMP, definition, npoph
from pyontutils.namespaces import rdfs, rdf, owl
from pyontutils.utils import allMembers
from neurondm import NeuronCUT, Config, Phenotype, NegPhenotype, LogicalPhenotype
from neurondm import EntailedPhenotype
from neurondm.models.cuts import make_cut_id, fixname
from neurondm.core import log, OntId, OntTerm

try:
    breakpoint
except NameError:
    from IPython import embed as breakpoint

def normalizeDoi(doi):
    if 'http' in doi:
        doi = '10.' + doi.split('.org/10.', 1)[-1]
    elif doi.startswith('doi:'):
        doi = doi.strip('doi:')
    elif doi.startswith('DOI:'):
        doi = doi.strip('DOI:')
    return doi


def select_by_curie_rank(results):
    ranking = 'CHEBI', 'UBERON', 'PR', 'NCBIGene', 'NCBITaxon', 'GO', 'SAO', 'NLXMOL'
    def key(result):
        if 'curie' in result:
            curie = result['curie']
        else:
            return 3, len(results) * 3

        prefix, _ = curie.split(':')
        if prefix in ranking:
            try:
                return 0, ranking.index(prefix)
            except ValueError:
                return 1, len(results) + 1
        else:
            return 2, len(results) * 2

    return sorted(results, key=key)[0]


def process_note(raw_note):
    if raw_note is None:
        return None
    p = ilxtr.literatureCitation
    for bit in  (b.strip() for b in raw_note.split('\n') if b.strip()):
        maybe_hypothesis = idFromShareLink(bit)
        if maybe_hypothesis:
            # TODO getDocInfoFromHypothesisId(maybe_hypothesis)
            yield p, rdflib.URIRef(shareLinkFromId(maybe_hypothesis))
        elif 'doi:' in bit or 'DOI:' in bit or 'doi.org' in bit:
            yield p, rdflib.URIRef('https://doi.org/' + normalizeDoi(bit))
        elif bit.startswith('http'):  # TODO parse the other things
            yield p, rdflib.URIRef(bit)
        else:
            yield p, rdflib.Literal(bit)  # FIXME cull editorial notes


class Cuts(Sheet):
    name = 'neurons-cut'


class CutsV1(Cuts):
    sheet_name = 'CUT V1.0'
    fetch_grid = True

    lnlu = {v:k for k, v in LogicalPhenotype.local_names.items()}

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'existing'):
            e_config = Config('cut-development')
            e_config.load_existing()
            # FIXME clear use case for the remaining bound to whatever query produced it rather
            # than the other way around ... how to support this use case ...
            # FIXME this ignores the sheet id mapping, and pulls in old/alternate axioms
            cls.existing = {n.origLabel.toPython():n for n in e_config.existing_pes}
            cls.existing.update({n.id_:n for n in e_config.existing_pes})
            cls.query = oq.OntQuery(oq.plugin.get('rdflib')(e_config.core_graph), instrumented=OntTerm)
            cls.sgv = Vocabulary()

        return super().__new__(cls)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def neurons(self, expect_pes):
        # TODO import existing ids to register by label
        self.config = Config('cut-roundtrip')
        self.skip = 'alignment label',
        self.errors = []
        self.failed = {}
        self.new = []
        self.release = []
        headers, *rows = self.values
        self.tomqc_check_ind = headers.index('tomqc')
        for i, neuron_row in enumerate(rows):
            yield from self.convert_row(i, neuron_row, headers, expect_pes)

    def convert_row(self, i, neuron_row, headers, expect_pes):
        id = None
        label_neuron  = None
        definition_neuron = None
        synonyms_neuron = None
        current_neuron = None
        phenotypes = []
        do_release = False
        predicate_notes = {}
        object_notes = {}
        other_notes = {}
        wat = {}
        def loop_internal(j, header, cell):
            nonlocal id
            nonlocal current_neuron
            nonlocal do_release
            notes = list(process_note(get_note(i + 1, j, self.cells_index)))  # + 1 since headers is removed
            if notes and not header.startswith('has'):
                _predicate = self.convert_other(header)
                if cell:
                    _object = rdflib.Literal(cell)  # FIXME curies etc.
                else:
                    _object = rdf.nil
                other_notes[_predicate, _object] = notes

            if header == 'curie':
                id = OntId(cell).u if cell else None
                return
            elif header == 'label':
                if id == OntId('NIFEXT:66').u:
                    breakpoint()
                label_neuron = cell
                if cell in self.existing:
                    current_neuron = self.existing[cell]
                elif cell:
                    # TODO
                    self.new.append(cell)
                else:
                    raise ValueError(cell)  # wat
                return
            elif header == 'Status 1.0':
                # TODO
                if cell == 'Yes':
                    do_release = True
                elif cell == 'Maybe':
                    pass
                elif cell == 'Not yet':
                    pass
                elif cell == 'Delete':
                    pass
                else:
                    pass

                return
            elif header == 'PMID':
                # TODO
                return
            elif header == 'Other reference':
                # TODO
                return
            elif header == 'Other label':
                # TODO
                return
            elif header == 'definition':
                return  # FIXME single space differences between the spreadsheet and the source

                if cell:
                    definition_neuron = rdflib.Literal(cell)

            elif header == 'synonyms':
                if cell:
                    synonyms_neuron = [rdflib.Literal(s.strip())
                                    # FIXME bare comma is extremely dangerous
                                    for s in cell.split(',')]

                return
            elif header in self.skip:
                return

            objects = []
            if cell:
                predicate = self.convert_header(header)
                if predicate is None:
                    log.debug(f'{(header, cell, notes)}')

                for object, label in self.convert_cell(cell):
                    if predicate in NeuronCUT._molecular_predicates:
                        if isinstance(object, tuple):
                            op, *rest = object
                            rest = [OntTerm(o).asIndicator().URIRef for o in rest]
                            object = op, *rest
                        elif object:
                            log.debug(f'{object!r}')
                            object = OntTerm(object).asIndicator().URIRef

                    if isinstance(label, tuple):  # LogicalPhenotype case
                        _err = []
                        for l in label:
                            if self.lower_check(l, cell):
                                _err.append((cell, label))
                        if _err:
                            self.errors.extend(_err)
                        else:
                            objects.append(object)
                    elif self.lower_check(label, cell):
                        self.errors.append((cell, label))
                    elif str(id) == object:
                        self.errors.append((header, cell, object, label))
                        object = None
                    else:
                        objects.append(object)

                if notes:
                    # FIXME this is a hack to only attach to the last value
                    # since we can't distinguish at the moment
                    wat[predicate, object] = notes
                    if object is not None:
                        # object aka iri can be none if we don't find anything
                        object_notes[object] = notes
                    else:
                        predicate_notes[predicate] = notes
                        # FIXME it might also be simpler in some cases
                        # to have this be object_notes[object] = notes
                        # because we are much less likely to have the same
                        # phenotype appear attached to the different dimensions

                        # FIXME comma sep is weak here because the
                        # reference is technically ambiguous
                        # might be an argument for the denormalized form ...
                        # or perhaps having another sheet for cases like that

            else:
                return

            if predicate and objects:
                for object in objects:  # FIXME has layer location phenotype
                    if isinstance(object, tuple):
                        op, *rest = object
                        pes = (Phenotype(r, predicate) for r in rest)  # FIXME nonhomogenous phenotypes
                        phenotypes.append(LogicalPhenotype(op, *pes))
                    elif object:
                        phenotypes.append(Phenotype(object, predicate))
                    else:
                        self.errors.append((object, predicate, cell))
            elif objects:
                self.errors.append((header, objects))
            else:
                self.errors.append((header, cell))
            # translate header -> predicate
            # translate cell value to ontology id

        #########################################

        for j, (header, cell) in enumerate(zip(headers, neuron_row)):
            loop_internal(j, header, cell)

        if current_neuron and phenotypes:
            # TODO merge current with changes
            # or maybe we just replace since all the phenotypes should be there?
            log.debug(phenotypes)
            if id is not None:
                log.debug(f'{(id, bool(id))}')

            elif label_neuron:
                id = make_cut_id(label_neuron)

            if id not in expect_pes:
                if id is not None:
                    log.error(f'{id!r} not in cuts!?')

                return

            phenotypes = sorted(set(phenotypes))
            _ep = expect_pes[id]
            if not allMembers(_ep, *phenotypes) and not neuron_row[self.tomqc_check_ind]:
            #if expect_pes[id] != len(phenotypes):
                # FIXME this is not a strict roundtrip, it may also include additions
                lp = len(phenotypes)
                lep = len(_ep)
                if lp == lep:
                    (pprint(sorted(_ep)))
                    (pprint(phenotypes))
                    pprint(set(_ep) - set(phenotypes))
                    pprint(set(phenotypes) - set(_ep))
                    _AAAAAA = id  # hack for debugger
                    print(_AAAAAA)
                log.error(f'{id!r} failed roundtrip {lp} != {lep}')
                self.failed[id] = phenotypes
                return

            neuron = NeuronCUT(*phenotypes, id_=id, label=label_neuron,
                               override=bool(id) or bool(label_neuron))
            neuron.adopt_meta(current_neuron)
            # FIXME occasionally this will error?!
            yield neuron

        else:
            return  # FIXME this polutes everything ???
        """
            fn = fixname(label_neuron)
            if not phenotypes and i:  # i skips header
                self.errors.append((i, neuron_row))  # TODO special review for phenos but not current
                phenotypes = Phenotype('TEMP:phenotype/' + fn),

            neuron = NeuronCUT(*phenotypes,
                               id_=make_cut_id(label_neuron),
                               label=label_neuron, override=True)

        """
        ###################################################

        # update the meta if there were any changes
        if definition_neuron is not None:
            neuron.definition = definition_neuron

        if synonyms_neuron is not None:
            neuron.synonyms = synonyms_neuron

        try:
            neuron.batchAnnotateByObject(object_notes)
            neuron.batchAnnotate(other_notes)
        except AttributeError as e:
            #breakpoint()
            log.exception(e) #'something very strage has happened\n', e)
            pass  # FIXME FIXME FIXME

        #neuron.batchAnnotateByPredicate(predicate_notes)  # TODO
        # FIXME doesn't quite work in this context, but there are other
        # cases where annotations to the general modality are still desireable
        # FIXME there may be no predicate? if the object fails to match?

        if do_release:
            self.release.append(neuron)


    @staticmethod
    def convert_header(header):
        if header.startswith('has'):  # FIXME use a closed namespace
            if header == 'hasLocationPhenotype':
                # can't have duplicate headers in the sheet so fix it here
                header = 'hasSomaLocatedIn'

            return npoph[header]
        else:
            return None

    @staticmethod
    def convert_other(header):
        if header == 'label':
            return rdfs.label
        elif header == 'curie':
            return rdf.type
        elif header == 'definition':
            return definition
        else:
            header = header.replace(' ', '_')
            return TEMP[header]  # FIXME

    @staticmethod
    def lower_check(label, cell):
        return label not in cell and label.lower() not in cell.lower()  # have to handle comma sep case

    _prefix_exclude = ('FMA',)
    _search_prefixes = ('UBERON', 'CHEBI', 'PR', 'NCBIGene', 'NCBITaxon',
                        'ilxtr', 'NIFEXT', 'SAO', 'NLXMOL', 'BIRNLEX', 'FMA', 'PATO')

    @classmethod
    def mapCell(cls, cell, syns=False, predicate=None):
        search_prefixes = tuple(p for p in cls._search_prefixes if p not in cls._prefix_exclude)

        if predicate and predicate in Phenotype._molecular_predicates:
            # uberon syns pollute molecular results so move it to one before birnlex
            ub, *rest, b = search_prefixes
            search_prefixes = (*rest, ub, b)

        if cell == 'contralateral':
            return ilxtr.Contralateral, cell  # XXX FIXME only BSPO has this right now
        elif cell.lower() == 'gaba receptor role':
            return ilxtr.GABAReceptor, cell

        if ':' in cell and ' ' not in cell:
            log.debug(cell)
            if 'http' in cell:
                if cell.startswith('http'):
                    t = OntTerm(iri=cell)
                else:
                    return None, None  # garbage with http inline
            else:
                t = OntTerm(cell, exclude_prefix=('FMA',))  # FIXME need better error message in ontquery

            return t.u, t.label

        if cell in ('Vertebrata', ):  # search syns
            syns = True

        def rank_mask(r):
            """
            create a boolean array testing if the current entry
            starts with the prefixes in order and what you will
            get out is arrays where the nth element is true if
            the nth prefix is matched which will then be sorted by n
            1 0 0 0 0 0 0 1 \\
            1 0 0 0 0 0 0 0 \\
            0 1 0 0 0 0 0 0 \\
            0 0 1 0 0 0 0 0 \\
            0 0 0 1 0 0 0 0 \\
            0 0 0 0 1 0 0 0 \\
            """
            # why did it take so long to think of this?
            return (
                *(r['curie'].startswith(p) for p in search_prefixes),
                'labels' in r and cell in r['labels'],
            )

        result = sorted([r for r in cls.sgv.findByTerm(cell, searchSynonyms=syns, prefix=search_prefixes)
                         if not r['deprecated']], key=rank_mask, reverse=True)
        #printD(cell, result)
        if not result:
            log.debug(f'{cell}')
            maybe = list(cls.query(label=cell, exclude_prefix=cls._prefix_exclude))
            if maybe:
                t = maybe[0]
                return t.u, t.label
            elif not syns:
                return cls.mapCell(cell, syns=True, predicate=predicate)
            else:
                return None, None
        elif len(result) > 1:
            #printD('WARNING', result)
            result = result[0] #select_by_curie_rank(result)
        else:
            result = result[0]

        return rdflib.URIRef(result['iri']), result['labels'][0]

    @classmethod
    def convert_cell(cls, cell_or_comma_sep, predicate=None):
        #printD('CONVERTING', cell_or_comma_sep)
        for cell_w_junk in cell_or_comma_sep.split(','):  # XXX WARNING need a way to alter people to this
            cell = cell_w_junk.strip()
            negative = (cell.lower().startswith('lacks') or
                        cell.lower().startswith('not '))
            if negative:
                __lacks, cell = cell.split(' ', 1)

            if cell.startswith('(OR') or cell.startswith('(AND'):
                start, *middle, end = cell.split('" "')
                OPoperator, first = start.split(' "')
                operator = OPoperator[1:]
                operator = cls.lnlu[operator]
                last, CP = end.rsplit('"')
                iris, labels = [], []
                for term in (first, *middle, last):
                    iri, label = cls.mapCell(term, predicate=predicate)
                    if label is None:
                        label = cell_or_comma_sep
                    if negative:
                        iri = LacksObject(iri)

                    iris.append(iri)
                    labels.append(label)

                yield (operator, *iris), tuple(labels)

            else:
                iri, label = cls.mapCell(cell, predicate=predicate)
                if negative:
                    iri = LacksObject(iri)

                if label is None:
                    yield iri, cell_or_comma_sep  # FIXME need a way to handle this that doesn't break things?
                else:
                    yield iri, label


class LacksObject(rdflib.URIRef):
    """ Objects that can indicate that they
        should be lifted to negative phenotypes

        This is a trick we can play to with the type system
        to pack the negative phenotype into a single object
        that could then be expanded to object.asPhenotype(predicate)

        potential alternate names:
        NegativeObject
        NegativePhenotypeValue
        AbsentObject
    """

    def asURIRef(self):
        return rdflib.URIRef(self)


class CutsV1Lite(Cuts):
    sheet_name = 'CUT V1.0'
    fetch_grid = False


class Row(sheets.Row):

    neuron_class = NeuronCUT

    entail_predicates = (
        ilxtr.hasAxonLocatedIn,
        ilxtr.hasDendriteLocatedIn,
        ilxtr.hasDendriteMorphologicalPhenotype,
    )

    def neuron_existing(self):
        curie = self.curie().value
        if curie:
            id_ = OntTerm(curie).u
            match = self.sheet.existing.get(id_)
            if match:
                match.id_ = id_
                return match

        al = self.alignment_label().value
        nrn = self.sheet.existing.get(al if al else self.label().value)

        if nrn and curie:
            nrn.id_ = id_

        return nrn

    def status(self):
        # column header renamed without warning
        return self.status_1_0()

    def include(self):
        return self.status().value == 'Yes'

    def entailed_molecular_phenotypes(self):
        cell = self.exhasmolecularphenotype()
        labels = cell.value.split(',')

        # FIXME hack
        yield OntTerm(curie='ilxtr:GABAReceptor').asPhenotype()
        yield OntTerm(curie='ilxtr:glutamateReceptor').asPhenotype()

        for label in labels:
            label = label.strip()
            term = self.sheet.sgv.findByTerm(label)
            if term:
                ot = OntTerm(iri=term[0]['iri'])
                p = ot.asPhenotype()
                yield p
                if ot.curie == 'NLXMOL:1006001':
                    yield OntTerm(curie='ilxtr:GABAReceptor').asPhenotype()
                elif ot.curie == 'SAO:1164727693':
                    yield OntTerm(curie='ilxtr:glutamateReceptor').asPhenotype()

    def asPhenotypes(self):
        log.warning(f'New neuron from sheet! {self.label().value}')
        has = [(attr, ilxtr[suffix])
               for attr, suffix in zip(self.header, self.sheet.values[0])
               if attr.startswith('has')]

        def map(attr, predicate):
            cell = getattr(self, attr)()
            value = cell.value
            if value:
                for iri, label in list(self.sheet.convert_cell(value, predicate=predicate)):
                    if ',' not in value and label != value:
                        log.warning(f'label mismatch {label!r} != {value!r}')

                    if iri is None:
                        if label == 'bed nucleus of stria terminalis juxtacapsular nucleus':
                            iri = OntTerm('UBERON:0011173', label='anterior division of bed nuclei of stria terminalis')
                        else:
                            log.debug(f'nothing found for {label}')
                            continue

                    if isinstance(iri, tuple):
                        op, *rest = iri  # TODO need combinators in future version for union/intersection of object
                        out = (op, *(NegPhenotype(r, predicate)
                                     if isinstance(r, LacksObject) else
                                     Phenotype(r, predicate)
                                     for r in rest if r is not None))
                        yield out
                        continue
                    elif isinstance(iri, LacksObject):
                        pheno = NegPhenotype(iri.asURIRef(), predicate)
                    else:
                        pheno = Phenotype(iri, predicate)

                    yield pheno.asIndicator()


        mapped = [(v, p) for attr, p in has for v in map(attr, p)]
        pes = [LogicalPhenotype(*value)
               if isinstance(value, tuple) else
               value
               #Phenotype(value, dimension)
               for value, dimension in mapped
               if value is not None]
        # moveToIndicator
        # exHasMolecularPhenotype
        # PMID
        # Other reference
        # Other label
        # definition
        # synonyms
        return pes

    def asNeuron(self):
        pes = self.asPhenotypes()
        return self.neuron_class(*pes, label=self.label().value)

    def _ec(self, nrn):
        # handle cases where we accidentally duplicated ids
        if hasattr(self, 'curie_extra'):
            ce = self.curie_extra().value
            if ce:
                eid = OntId(ce).u
                nrn.add_objects(owl.equivalentClass, eid)

    def neuron_cleaned(self, context=nullcontext()):
        # FIXME WARNING nullcontext behavior seems to have changed at some point !?
        conditional_entailed_predicates = (
            (ilxtr.hasSomaLocatedInLayer, (False or self.curie().value == 'NIFEXT:55')),  # martinotti cell
        )

        ne = self.neuron_existing()
        emp = list(self.entailed_molecular_phenotypes())
        eobjects = [e.p for e in emp]
        def should_entail(pe):
            return (pe.p in eobjects or
                    pe.e in self.entail_predicates or
                    (pe.e, True) in conditional_entailed_predicates or
                    isinstance(pe, LogicalPhenotype) and any(should_entail(_) for _ in pe.pes) or
                    # see this comment for discussion about negative phenotypes on CUTs
                    # https://github.com/SciCrunch/NIF-Ontology/issues/222#issuecomment-680175477
                    isinstance(pe, NegPhenotype))

        sheet_pes = self.asPhenotypes()
        sheet_pes = [pe.asEntailed() if should_entail(pe) else pe for pe in sheet_pes]

        if ne is None:
            curie = self.curie().value
            id_ = curie if curie else None
            nrn = self.neuron_class(*sheet_pes, id_=id_, label=self.label().value, override=id_ is None)
            self._ec(nrn)
            return nrn

        if not emp:
            # can't just return the existing neuron because it isn't bound to the current config
            # FIXME uh what an aweful design
            nrn = self.neuron_class(*ne, *sheet_pes, id_=ne.id_, label=ne.origLabel,
                                    override=True).adopt_meta(ne)
            self._ec(nrn)
            return nrn

        pes = [pe.asEntailed() if should_entail(pe) else pe for pe in ne]

        with context:
            nrn = self.neuron_class(*pes, *sheet_pes, id_=ne.id_, label=ne.origLabel, override=True)
            self._ec(nrn)
            return nrn


def main():
    #from neurondm.models.cuts import main as cuts_main
    #cuts_config, *_ = cuts_main()

    # monkey patch inside main to ensure that other code doesn't change while we are at the top level
    sheets.Row = Row

    from neurondm.compiled.common_usage_types import config as cuts_config
    cuts_neurons = cuts_config.neurons()
    expect_pes = {n.id_:n.pes for n in cuts_neurons}

    sheet = CutsV1()
    _neurons = list(sheet.neurons(expect_pes))
    config = sheet.config
    errors = sheet.errors
    new = sheet.new
    release = sheet.release

    #sheet.show_notes()
    config.write_python()
    config.write()
    #config = Config(config.name)
    #config.load_existing()  # FIXME this is a hack to get get a load_graph

    # FIXME we need this because _bagExisting doesn't deal with unionOf right now
    def trything(f):
        @wraps(f)
        def inner(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except:
                pass

        return inner

    from neurondm import Config, NeuronCUT

    failed_config = Config('cut-failed')
    [trything(NeuronCUT)(*pes, id_=id_) for id_, pes in sheet.failed.items()]
    failed_config.write_python()
    failed_config.write()

    release_config = Config('cut-release')
    [NeuronCUT(*n, id_=n.id_, label=n.origLabel, override=True).adopt_meta(n) for n in release]
    release_config.write_python()
    release_config.write()

    from neurondm.models.cuts import export_for_review
    review_rows = export_for_review(config, [], [], [], filename='cut-rt-test.csv', with_curies=True)
    from pyontutils.utils import byCol
    valuesC = byCol(sheet.values[1:],
                    header=[v.replace(' ', '_') for v in sheet.values[0]],
                    to_index=['label'])
    reviewC = byCol(review_rows[1:], header=[v.replace(' ', '_') for v in review_rows[0]], to_index=['label'])
    def grow(r):
        log.debug(r)
        # TODO implement on the object to allow joining on an index?
        # man this would be easier with sql >_< probably pandas too
        # but so many dependencies ... also diffing issues etc
        if r.label is not None:
            return valuesC.searchIndex('label', r.label)

    def key(field_value):
        field, value = field_value
        try:
            return 0, valuesC.header._fields.index(field)  # TODO warn on field mismatch
        except ValueError as e:
            log.error(f'{field} {value}')
            return 1, 0

    def replace(r, *cols):
        """ replace and reorder """
        # FIXME _super_ inefficient
        vrow = grow(r)
        log.debug('\n'.join(r._fields))
        log.debug('\n'.join(str(_) for _ in r))
        for field, value in sorted(zip(r._fields, r), key=key):
            if field in cols:
                value = getattr(vrow, field)

            yield '' if value is None else value  # completely overwrite the sheet

    breakpoint()
    rows = [list(replace(r, 'Status 1.0', 'definition', 'synonyms', 'PMID')) for r in reviewC]
    #resp = update_sheet_values('neurons-cut', 'Roundtrip', rows)
    if __name__ == '__main__':
        breakpoint()


def main():
    #cv1 = CutsV1Lite()
    sheets.Row = Row
    CutsV1.fetch_grid = False
    cv1 = CutsV1()
    hrm = [cv1.row_object(i) for i, r in enumerate(cv1.values)
           if cv1.row_object(i).exhasmolecularphenotype().value]
    to_sco = set(t for h in hrm for t in h.entailed_molecular_phenotypes())
    ros = [cv1.row_object(i + 1) for i, r in enumerate(cv1.values[1:])]
    to_fix = [r for r in ros if list(r.entailed_molecular_phenotypes())]
    #maybe_fixed = [t.neuron_cleaned() for t in to_fix]
    #assert maybe_fixed != [f.neuron_existing() for f in to_fix]
    config = Config('common-usage-types')
    _final = [r.neuron_cleaned() for r in ros if r.include()]
    final = [f for f in _final if f is not None]  # FIXME there are 16 neurons marked as yes that are missing
    #fixed = [f for f in final if [_ for _ in f.pes if isinstance(_, EntailedPhenotype)]]
    [f._sigh() for f in final]
    config.write()
    config.write_python()
    if __name__ == '__main__':
        breakpoint()


if __name__ == '__main__':
    log.setLevel('WARNING')
    main()



from test.common import _TestNeuronsBase, pyel, tel
'''
class TestNeurons(_TestNeuronsBase):
    # TODO make sure this runs after cli test? it should ...
    # but then we need to keep the output of ndl around

    def test_load_existing(self):
        from neurondm.lang import Neuron, Config
        config = Config('neuron_data_lifted')
        config.load_existing()
        assert len(Neuron.load_graph)
        neurons = Neuron.neurons()
        assert neurons

    def test_zz_adopt(self):
        from neurondm.lang import Neuron, Phenotype, Config
        ndl_config = Config('neuron_data_lifted')
        ndl_config.load_existing()
        bn_config = Config('basic-neurons')
        bn_config.load_existing()
        ndl_neurons = ndl_config.neurons()
        bn_neurons = bn_config.neurons()
        config = Config('__test_output', ttl_export_dir=tel)
        shapeshifter = Neuron(Phenotype('TEMP:soul-stealer'))
        for n in ndl_neurons:
            shapeshifter.adopt_meta(n)

        assert list(n.synonyms)
        assert list(n.definitions)
        n.write()

        assert list(shapeshifter.synonyms)
        assert list(shapeshifter.definitions)
        shapeshifter.write()

        # it eats _all_ of them
        allsyns = set(shapeshifter.synonyms)
        alldefs = set(shapeshifter.definitions)
        assert all(s in allsyns for s in n.synonyms)
        assert all(s in alldefs for s in n.definitions)

    def test_fail(self):
        from neurondm.lang import Neuron, Phenotype, Config
        bn_config = Config('basic-neurons')
        # TODO config.activate()? context manager for config ... too ...
        Neuron(Phenotype('TEMP:test'))
        try:
            bn_config.load_existing()
            raise AssertionError('Should have failed because a neuron has been created')
        except Config.ExistingNeuronsError as e:
            pass

    def test_0_equality(self):
        """ make sure that __eq__ and __hash__ are implmented correctly """
        from neurondm import Config, Neuron, Phenotype, NegPhenotype
        from neurondm import LogicalPhenotype, AND, OR, ilxtr
        from neurondm import NeuronCUT, NeuronEBM

        config = Config('test-equality', ttl_export_dir=tel, py_export_dir=pyel)

        pp = Phenotype(ilxtr.a)
        pn = NegPhenotype(ilxtr.a)
        assert pp != pn
        assert hash(pp) != hash(pn)

        assert Phenotype(ilxtr.c) == Phenotype(ilxtr.c)
        assert NegPhenotype(ilxtr.c) == NegPhenotype(ilxtr.c)
        assert hash(Phenotype(ilxtr.c)) == hash(Phenotype(ilxtr.c))
        assert hash(NegPhenotype(ilxtr.c)) == hash(NegPhenotype(ilxtr.c))

        lppo = LogicalPhenotype(OR, pp, Phenotype(ilxtr.b))
        lppa = LogicalPhenotype(AND, pp, Phenotype(ilxtr.b))
        lpno = LogicalPhenotype(OR, pn, NegPhenotype(ilxtr.b))
        assert lppo != lppa
        assert lppo != lpno
        assert lppa != lpno
        assert hash(lppo) != hash(lppa)
        assert hash(lppo) != hash(lpno)
        assert hash(lppa) != hash(lpno)

        lpe1 = LogicalPhenotype(OR, Phenotype(ilxtr.c), Phenotype(ilxtr.d))
        lpe2 = LogicalPhenotype(OR, Phenotype(ilxtr.c), Phenotype(ilxtr.d))
        assert lpe1 == lpe2
        assert hash(lpe1) == hash(lpe2)

        # make sure that self.pes equality works
        assert (pp, pn, lppo, lppa, lpno) == (pp, pn, lppo, lppa, lpno)

        npp = Neuron(pp)
        npn = Neuron(pn)
        assert npp != npn
        assert hash(npp) != hash(npn)

        nlppo = Neuron(lppo)
        nlppa = Neuron(lppa)
        nlpno = Neuron(lpno)

        assert nlppo != nlppa
        assert nlppo != nlpno
        assert nlppa != nlpno
        assert hash(nlppo) != hash(nlppa)
        assert hash(nlppo) != hash(nlpno)
        assert hash(nlppa) != hash(nlpno)

        assert Neuron(pp) == Neuron(pp)
        assert Neuron(pn) == Neuron(pn)
        assert hash(Neuron(pp)) == hash(Neuron(pp))
        assert hash(Neuron(pn)) == hash(Neuron(pn))

        assert NeuronCUT(pp) != NeuronEBM(pp)
        assert hash(NeuronCUT(pp)) != hash(NeuronEBM(pp))


'''

class TestRoundtrip(_TestNeuronsBase):
    # need to test other more complex constructs

    pyname = 'test-py'
    ttlname = 'test-ttl'

    def setUp(self):
        super().setUp()
        from neurondm import Config, Neuron, Phenotype, NegPhenotype
        from neurondm import EntailedPhenotype
        self.Config = Config
        self.Neuron = Neuron
        self.Phenotype = Phenotype
        self.NegPhenotype = NegPhenotype
        self.EntailedPhenotype = EntailedPhenotype

    def tearDown(self):
        super().tearDown()

    def test_py_simple(self):

        config = self.Config(self.pyname, ttl_export_dir=tel, py_export_dir=pyel)
        n1 = self.Neuron(self.Phenotype('TEMP:python-phenotype'))
        n2 = self.Neuron(self.NegPhenotype('TEMP:python-phenotype'))
        assert n1 != n2
        self.Neuron(self.Phenotype('TEMP:python-location', 'ilxtr:hasLocationPhenotype'))
        self.Neuron(self.NegPhenotype('TEMP:python-location', 'ilxtr:hasLocationPhenotype'))
        config.write_python()

        config2 = self.Config(self.pyname, ttl_export_dir=tel, py_export_dir=pyel)
        config2.load_python()  # FIXME load existing python ...
        config2.write()

        config2.out_graph.debug()

        config3 = self.Config(self.pyname, ttl_export_dir=tel, py_export_dir=pyel)
        config3.load_existing()

        assert config.existing_pes is not config2.existing_pes is not config3.existing_pes
        assert config.neurons() == config2.neurons() == config3.neurons()

    def test_ttl_simple(self):
        # madness spreads apparently, here is a minimal repro for the issue
        # pytest test/test_madness.py test/test_neurons.py -k 'test_ttl_simple or test_entailed_predicate'
        # The other classes in this file can be commented out
        # an even more specific repro
        # pytest test/test_madness.py test/test_neurons.py \
        # -k 'test_madness and test_ttl_simple or
        #     test_neurons and test_entailed_predicate or
        #     test_neurons and test_ttl_simple'

        config = self.Config(self.ttlname, ttl_export_dir=tel, py_export_dir=pyel)
        self.Neuron(self.Phenotype('TEMP:turtle-phenotype'))
        self.Neuron(self.NegPhenotype('TEMP:turtle-phenotype'))
        self.Neuron(self.Phenotype('TEMP:turtle-location', 'ilxtr:hasLocationPhenotype'))
        self.Neuron(self.NegPhenotype('TEMP:turtle-location', 'ilxtr:hasLocationPhenotype'))
        config.write()
        a = config.neurons()

        config2 = self.Config(self.ttlname, ttl_export_dir=tel, py_export_dir=pyel)
        config2.load_existing()
        config2.write_python()
        b = config2.neurons()

        config3 = self.Config(self.ttlname, ttl_export_dir=tel, py_export_dir=pyel)
        config3.load_python()
        c = config3.neurons()

        print(a, b, c)
        assert config.existing_pes is not config2.existing_pes is not config3.existing_pes
        if not a == b == c:
            breakpoint()
        # so somehow when test_entailed_predicate is called along with test_ttl_simple
        # n1 from that sneeks into config3, but ONLY when this class is imported into
        # another file AND that file is run, so this seems like it is happening because
        # somehow the tep neuron persists through the tearDown, and for some reason
        # importing a testing module into another file is sufficient to keep the
        # garbage collector from collecting between runs or something ??!?
        assert a == b == c

    def test_entailed_predicate(self):
        p1 = self.Phenotype('ilxtr:somewhere', 'ilxtr:hasLocationPhenotype')
        p2 = self.EntailedPhenotype('ilxtr:somewhere', 'ilxtr:hasLocationPhenotype')
        n1 = self.Neuron(p1, p2)
        n1._graphify()
        # TODO assert to make sure the pattern is right


class TestRoundtripCUT(TestRoundtrip):

    pyname = 'test-cut-py'  # can't import the same module name twice
    ttlname = 'test-cut-ttl'

    def setUp(self):
        super().setUp()
        from neurondm import NeuronCUT
        self.Neuron = NeuronCUT



class TestLabels(_TestNeuronsBase):
    def setUp(self):
        super().setUp()
        from neurondm import Config, Neuron, Phenotype, NegPhenotype
        from neurondm import EntailedPhenotype, LogicalPhenotype, AND
        self.AND = AND
        self.Config = Config
        self.Neuron = Neuron
        self.Phenotype = Phenotype
        self.NegPhenotype = NegPhenotype
        self.EntailedPhenotype = EntailedPhenotype
        self.LogicalPhenotype = LogicalPhenotype

    def test_nest_logical(self):
        AND = self.AND
        Neuron = self.Neuron
        LogicalPhenotype = self.LogicalPhenotype
        Phenotype = self.Phenotype
        n1 = Neuron(Phenotype('NCBITaxon:10090',
                              'ilxtr:hasInstanceInTaxon',
                              label='Mus musculus'),
                    LogicalPhenotype(AND,
                                     Phenotype('NCBIGene:50779',
                                               'ilxtr:hasExpressionPhenotype',
                                               label='Rgs6'),
                                     LogicalPhenotype(AND,
                                                      Phenotype('ilxtr:GABAReceptor',
                                                                'ilxtr:hasExpressionPhenotype',
                                                                label='GABAR'))),
                    label='test logical')

        n2 = Neuron(Phenotype('NCBITaxon:10090',
                              'ilxtr:hasInstanceInTaxon',
                              label='Mus musculus'),
                    LogicalPhenotype(AND,
                                     Phenotype('NCBIGene:50779',
                                               'ilxtr:hasExpressionPhenotype',
                                               label='Rgs6'),
                                     Phenotype('ilxtr:GABAReceptor',
                                               'ilxtr:hasExpressionPhenotype',
                                               label='GABAR')),
                    label='test logical')

        n1l = n1.genLabel
        n2l = n2.genLabel
        assert '(intersectionOf' in n1l, n1l
        assert '(intersectionOf' in n2l, n2l



#!/usr/bin/env python3

from neurondm.sheets import Sheet
from neurondm import OntId, OntTerm, Config, NeuronEBM, Neuron
from pyontutils.utils import byCol, relative_path
from pyontutils.namespaces import ilxtr
from pyontutils.closed_namespaces import rdfs


class APIN(Sheet):
    name = 'apinatomy-to-npo'
    sheet_name = 'Bolser-Lewis EBM'

    @staticmethod
    def convert(value):
        if value == 'Projection neuron':
            e = ilxtr.ProjectionPhenotype
        else:
            e = value

        id = OntId(e.strip()) if ':' in e else e.strip()
        if id.prefix == 'ILX':
            t = id.asInstrumented()
            id = t.asPreferred()

        return id

    @property
    def reduced(self):
        et = tuple()
        header = self.values[0]
        skip = [header[i] for i, (h, *rest) in enumerate(self.byCol.cols) if not any(rest)]
        for row in self.values[1:31]:
            if not any(row):
                continue

            stuff = []
            for h, column in zip(header, row):
                if h in skip:
                    continue

                predicate = ilxtr[h] if h != 'label' else rdfs.label
                if column:
                    if h == 'label':
                        objects = [column]
                    else:
                        objects = [self.convert(e) for e in column.split(',')]
                else:
                    objects = et
                
                stuff.append((predicate, objects))

            yield stuff

    @property
    def bags(self):
        for row in self.reduced:
            phenotypes = []
            label = None
            id = None
            for predicate, objects in row:
                if predicate == rdfs.label:
                    if objects:
                        label = objects[0]

                    continue
                elif predicate == ilxtr.curie:
                    if objects:
                        id = objects[0]

                    continue

                for o in objects:
                    pheno = OntTerm(o).asPhenotype(predicate=predicate)
                    phenotypes.append(pheno)

            yield id, label, phenotypes


class BolserLewisNeuron(NeuronEBM):
    owlClass = ilxtr.NeuronBolserLewis
    shortname = 'bolew'
    

def main():
    a = APIN()
    config = Config('bolser-lewis',
                    source_file=relative_path(__file__, no_wd_value=__file__))
    bags = list(a.bags)
    for id, label, bag in bags:
        BolserLewisNeuron(*bag, label=label, id_=id, override=True)

    config.write()
    labels = (
        rdfs.label,
        #ilxtr.genLabel, ilxtr.localLabel, ilxtr.simpleLabel,
        #ilxtr.simpleLocalLabel, skos.prefLabel
    )
    to_remove = [t for t in config._written_graph
                 if t[1] in labels]
    [config._written_graph.remove(t) for t in to_remove]
    config._written_graph.write()
    config.write_python()
    return config,


if __name__ == '__main__':
    main()



def nst_to_adj(l):
    start = l[0]
    if (isinstance(start, list) or isinstance(start, tuple)):
        raise ValueError('malformed')

    first = [] if start == "blank" else [(start, e[0]) for e in l[1:] if e]
    second = [p for r in l[1:] if r for p in nst_to_adj(r)]
    return tuple(first + second)

def assoc(k, ass):
    for a, b in ass:
        if a == k:
            return a, b

def dvals(k, ass):
    pair = assoc(k, ass)
    mem = ass[ass.index(pair):] if pair in ass else False
    tail = mem[1:] if mem else tuple()
    yield from ((pair[1], *dvals(k, tail)) if tail else pair[1:]) if pair else tuple()


def inest(adj, start, seen):
    nl = tuple()
    out = []
    maybe_out = []
    seen = {start} | seen
    for next_start in dvals(start, adj):
        if next_start in seen:
            # do not append if next_start already in seen
            nl, snext = (next_start,), seen
            if not out:  # multi-parent case
                # maybe out can already have results due to multi-child case
                maybe_out.append(nl)
        else:
            nl, snext = inest(adj, next_start, seen)
            out.append(nl)

        seen = snext

    if not out and maybe_out:
        out = maybe_out

    l = tuple(out)
    return ((start, *l) if l else (start,)), seen


def adj_to_nst(adj, start=None):
    keys = set([a[0] for a in adj])
    values = set([a[1] for a in adj])
    inverted = [(a[1], a[0]) for a in adj]
    starts = keys - values
    ends = values - keys

    seen = set()
    nl = None
    out = []
    for start in starts:
        nl, snext = inest(adj, start, seen)
        seen = snext
        out.append(nl)

    l = tuple(out)
    if len(l) == 1 and (isinstance(l[0], list) or isinstance(l[0], tuple)):
        return l[0]
    else:
        return ["blank", *l]


def lin_to_adj(linear):
    return tuple(zip(linear[:-1], linear[1:]))


def adj_to_lin(adj):
    # this is not deterministic when there are runs of equal length
    # this should return a list of linear paths from longest to
    # shortest that are sufficient to cover all nodes, essentially
    # the longest single path through the tree and then branches
    raise NotImplementedError('TODO')


def bind_rdflib():
    import rdflib
    from pyontutils.namespaces import rdf

    def to_rdf(g, nested):
        if isinstance(nested, list) or isinstance(nested, tuple):
            bn0 = rdflib.BNode()
            rdflib.collection.Collection(g, bn0, (to_rdf(g, n) for n in nested))
            return bn0
        elif isinstance(nested, rl):
            bn1 = rdflib.BNode()
            return nested.to_rdf(to_rdf, g, bn1)
        elif isinstance(nested, rdflib.term.Node):
            return nested
        else:
            return rdflib.Literal(nested)

    def from_rdf(g, linker, to_python=False):
        nested = {}
        for s, o in g[:linker:]:
            l = tuple(getlist(g, o, to_python=to_python))
            nested[s] = l

        return nested

    def getlist(g, bn0, to_python=False):
        def f(node, g):
            o = None
            for o in g[node:rdf.first:]:
                if isinstance(o, rdflib.BNode):
                    # unfortunately we have to branch on the result type
                    # or have to look ahead to see if the list continues
                    # picked the look ahead so that we don't have to check
                    # length and we don't have to know about the rl type
                    if list(g[o:rdf.first]):  # list continues
                        # XXX watch out for malformed lists that might not
                        # have a rdf:first ?
                        yield tuple(getlist(g, o, to_python=to_python))
                    else:  # likely [ :a :b ] case
                        yield from getlist(g, o, to_python=to_python)

                else:
                    if to_python and isinstance(o, rdflib.Literal):
                        yield o.toPython()
                    else:
                        yield o

            for o in g[node:rdf.rest:]:
                if o != rdf.nil:
                    yield from getlist(g, o, to_python=to_python)

            if o is None and isinstance(node, rdflib.BNode):  # [ :a  :b ] case
                for p, o in g[node:]:
                    if to_python:
                        yield rl((p.toPython() if isinstance(p, rdflib.Literal) else p),
                                 (o.toPython() if isinstance(o, rdflib.Literal) else o),)
                    else:
                        yield rl(p, o)

        yield from g.transitiveClosure(f, bn0)

    return to_rdf, from_rdf


class rl:
    def __init__(self, region, layer=None):
        self.region = region
        self.layer = layer

    def __repr__(self):
        if self.layer:
            return f"{self.__class__.__name__}({self.region!r}, {self.layer!r})"
        else:
            return f"{self.__class__.__name__}({self.region!r})"

    def to_rdf(self, to_rdf, g, bn):
        r = to_rdf(g, self.region)
        if self.layer is None:
            return r

        l = to_rdf(g, self.layer)
        g.add((bn, r, l))
        return bn

    def __eq__(self, other):
        return type(self) == type(other) and self.region == other.region and self.layer == self.layer

    def __hash__(self):
        return hash((self.__class__, self.region, self.layer))

    def __lt__(self, other):
        # FIXME unstable
        return (type(self) == type(other) and
                type(self.region) == type(other.region) and
                self.region is not None and
                self.region < other.region and
                type(self.layer) == type(self.layer) and
                self.layer is not None and
                self.layer < self.layer)


def test():
    from pprint import pprint
    to_rdf, from_rdf = bind_rdflib()
    adj_test = (
        ("a", "b"),
        ("b", "c"),
        ("c", "d"),
        ("d", "e"),
        ("d", "k"),

        ("a", "f"),
        ("f", "g"),

        ("h", "i"),
        ("i", "j"),
    )

    nested = adj_to_nst(adj_test)
    print('nested', nested)
    print('asdf', list(dvals('a', adj_test)))
    print('asdf', list(dvals('c', adj_test)))
    print('asdf', list(dvals('d', adj_test)))

    adj_out = nst_to_adj(nested)
    print(adj_out)

    assert set(adj_test) == set(adj_out), 'oops'

    from pyontutils.core import OntGraph
    from pyontutils.namespaces import ilxtr, rdf

    g = OntGraph()


    bn = to_rdf(g, nested)
    g.add((ilxtr['sub-1'], ilxtr.predicate, bn))

    adj_2 = (
        (ilxtr.a, ilxtr.b),
        (ilxtr.b, ilxtr.c),
        (ilxtr.c, ilxtr.d),
        (ilxtr.d, ilxtr.e),
        (ilxtr.d, ilxtr.h),

        (ilxtr.a, ilxtr.f),
        (ilxtr.f, ilxtr.g),
    )
    nst_2 = adj_to_nst(adj_2)
    print('nst_2', nst_2)
    bn = to_rdf(g, nst_2)
    g.add((ilxtr['sub-2'], ilxtr.predicate, bn))

    lin_3 = [1, 2, 3, 4, 5, 6]
    adj_3 = lin_to_adj(lin_3)
    nst_3 = adj_to_nst(adj_3)
    print('nst_3', nst_3)
    bn = to_rdf(g, nst_3)
    g.add((ilxtr['sub-3'], ilxtr.predicate, bn))

    # multiple linear should work if we concat
    # the outputs together
    adj_4 = tuple(set([pair for lin in
                       ([1, 2, 3, 4, 5],
                        [3, 6, 7],)
             for pair in lin_to_adj(lin)]))
    nst_4 = adj_to_nst(adj_4)
    print('nst_4', nst_4)
    bn = to_rdf(g, nst_4)
    g.add((ilxtr['sub-4'], ilxtr.predicate, bn))

    adj_5 = (
        (rl(ilxtr.a,), rl(ilxtr.b,)),
        (rl(ilxtr.b,), rl(ilxtr.c, ilxtr.l1)),
        #(rl(ilxtr.c, ilxtr.l1), rl(ilxtr.d,)),  # alt end # this one does not duplicate the final node in adj_to_nst
        (rl(ilxtr.c, ilxtr.l1), rl(ilxtr.c, ilxtr.l2)),
        (rl(ilxtr.c, ilxtr.l2), rl(ilxtr.c, ilxtr.l3)),
        (rl(ilxtr.c, ilxtr.l3), rl(ilxtr.d,)),
    )
    nst_5 = adj_to_nst(adj_5)
    pprint(('nst_5', nst_5))
    bn = to_rdf(g, nst_5)
    g.add((ilxtr['sub-5'], ilxtr.predicate, bn))

    # multi-parent multi-child case
    adj_6 = (
        (3, 1),
        (3, 2),
        (4, 1),
        (4, 2),
        (5, 1),
        (5, 2),
        (6, 1),
        (6, 2),
        (1, 7),
        (1, 8),
        (2, 7),
        (2, 8),
    )
    nst_6 = adj_to_nst(adj_6)
    rej_6 = nst_to_adj(nst_6)
    pprint(('nst_6', nst_6))
    pprint(('rej_6', rej_6))
    assert sorted(adj_6) == sorted(rej_6)
    bn = to_rdf(g, nst_6)
    g.add((ilxtr['sub-6'], ilxtr.predicate, bn))

    un_rdf = from_rdf(g, ilxtr.predicate, to_python=True)
    assert un_rdf[ilxtr['sub-1']] == tuple(nested)
    assert un_rdf[ilxtr['sub-2']] == tuple(nst_2)
    assert un_rdf[ilxtr['sub-3']] == tuple(nst_3)
    assert un_rdf[ilxtr['sub-4']] == tuple(nst_4)
    un_rdf[ilxtr['sub-5']] == tuple(nst_5)  # non-invertable case due to layer=None ambiguity
    assert un_rdf[ilxtr['sub-6']] == tuple(nst_6)

    g.debug()


if __name__ == '__main__':
    test()



from pyontutils import sheets
from pyontutils.sheets import Sheet
from pyontutils.namespaces import ilxtr
from neurondm.sheets import CutsV1, Row as RowBase
from neurondm.core import Config, NeuronEBM
from neurondm.phenotype_namespaces import Species


class NeuronNerves(NeuronEBM):
    owlClass = ilxtr.NeuronNerves
    shortname = 'nerves'


class Row(RowBase):

    entail_predicates = tuple()
    neuron_class = NeuronNerves

    def include(self):
        return True

    def neuron_existing(self):
        return None

    def entailed_molecular_phenotypes(self):
        return []


class FromNLP(Sheet):
    name = 'nlp-pns'


class NervesEBM(CutsV1, FromNLP):
    name = 'nlp-pns'
    sheet_name = 'evidence-based modeling statements'
    fetch_grid = False

    _prefix_exclude = tuple()


def main():

    sheets.Row = Row  # monkey patch

    ner = NervesEBM()
    ros = [ner.row_object(i + 1) for i, r in enumerate(ner.values[1:])]
    config = Config('nerves')
    _final = [r.neuron_cleaned(context=NeuronNerves(Species.Human)) for r in ros if r.include()]
    # FIXME TODO add references
    final = _final
    [f._sigh() for f in final]
    config.write()
    config.write_python()
    neurons = config.neurons()
    n = neurons[0]
    return config,


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import rdflib
from pyontutils.namespaces import ilxtr, rdfs
from neurondm.core import LocalNameManager
from neurondm.lang import Phenotype, EntailedPhenotype, Neuron, NeuronEBM, Config

slp = 'ilxtr:hasSomaLocatedIn'
sllp = 'ilxtr:hasSomaLocatedInLayer'
alp = 'ilxtr:hasAxonLocatedIn'
dlp = 'ilxtr:hasDendriteLocatedIn'
synp = 'ilxtr:hasAxonPresynapticElementIn'
snsp = 'ilxtr:hasDendriteSensorySubcellularElementIn'  # XXX new, kind of like axon terminal but for dendrites

fconp = 'ilxtr:hasForwardConnectionPhenotype'

ntkb = rdflib.Namespace(ilxtr[''] + 'neuron-type-keast-')
ntku = rdflib.Namespace(ilxtr[''] + 'neuron-type-keast-unbranched-')
ntkh = rdflib.Namespace(ilxtr[''] + 'neuron-type-helper-keast-')
npkb = rdflib.Namespace(ilxtr[''] + 'neuron-phenotype-keast-')


class NeuronKeast2020(Neuron):  # FIXME should be an EBM but something is a bit off
    owlClass = 'ilxtr:NeuronKeast2020'
    shortname = 'Keast2020'


kNeuron = NeuronKeast2020


def ambig(name):
    return f'{name} (kblad)'


labels = {
1: ambig('pelvic ganglion parasympathetic neuron'),
2: ambig('pelvic ganglion sympathetic neuron'),
3: ambig('inferior mesenteric ganglion neuron'),
4: ambig('sympathetic chain ganglion neuron'),
5: ambig('parasympathetic spinal preganglionic neuron'),
6: ambig('sympathetic preganglionic neuron innervating pelvic ganglion neuron'),
7: ambig('sympathetic preganglionic neuron innervating inferior mesenteric ganglion neuron'),
8: ambig('sympathetic preganglionic neuron innervating sympathetic chain ganglion neuron'),
9: 'urethral rhabdosphincter motor neuron',
10: 'L6-S1 sensory neuron innervating bladder',
11: 'L1-L2 sensory neuron innervating bladder',
12: 'sensory neuron innervating urethral rhabdosphincter',}


def kl(n):
    if n in labels:
        return labels[n]

    return f'Neuron population {n} of the Keast bladder model'


def kd(n):
    return ('The neuron type corresponding to all the members of '
            f'neuron population {n} from the keast bladder model.')


def kld(n):
    return {'label': kl(n),
            'definition': kd(n),}

def needs_keast_namespace(config):
    """ define neurons for keast spinal """

    plut = {v:k for k, v in Keast2020.items()}
    def fn(phen):
        return plut[phen]

    # define the inferred parents for neuron populations
    for n in (2, 4, 5, 6, 7, 8, 9, 10, 11, 12):
        NeuronKeast2020(Keast2020[f'ntk_{n}'], id_=ntkb[str(n)], **kld(n))
        # FIXME hack
        config.out_graph.add((npkb[str(n)], rdfs.subClassOf, ilxtr.Phenotype))

    # FIXME spinal cord white matter axons are missing from these bags

    # neuron populations
    neuron_1 = [kNeuron(PG, n_bl, BDWsyn, BNWsyn, para, post, id_=ntkb['1'], **kld(1))]  # pos
    neuron_2 = [kNeuron(PG, n_bl, synloc, sym, post, ntk_2_ent, id_=ntku[f'2-{fn(synloc)}'])
                for synloc in (BDWsyn, BNWsyn)]  # sos # FIXME smooth muslce only
    neuron_3 = [kNeuron(IMG, BDWsyn, BNWsyn, sym, post, id_=ntkb['3'], **kld(3))]  # sos # FIXME smooth muscle only

    neuron_4 = []  # sos
    with Neuron(n_ps, PGax, n_bl, sym, post, id_=ntkh['common-4']):
        for soma_location_id in four_soma_locs:
            somaloc = Phenotype(soma_location_id, slp)
            for exits_via in ((L6_gr,), (S1ax, S1_gr,)):
                for synapses_on in (BNVWsyn, BDVWsyn):
                    # 4
                    n4 = kNeuron(
                        somaloc, *exits_via, synapses_on, ntk_4_ent,
                        id_=ntku[f'4-{fn(somaloc)}-{fn(exits_via[0])}-{fn(synapses_on)}'])
                    neuron_4.append(n4)

    neuron_5 = []  # pre
    with Neuron(VII, n_ps, PGax, PGsyn, ntk_1_fcon, para, pre,
                id_=ntkh['common-5']):
        for somaloc, ventral_root_exit in zip((L6,    S1),
                                              (L6_vr, S1_vr)):
            n5 = kNeuron(somaloc, ventral_root_exit, ntk_5_ent,
                         id_=ntku[f'5-{fn(somaloc)}'])
            neuron_5.append(n5)

    common = {L1: (L1_vr, L1_wr),
              L2: (L2_vr, L2_wr),}
    common67 = {L1: (L1_gr,),
                L2: (L2_gr,),}

    neuron_6 = []  # sre
    neuron_7 = []  # sre
    neuron_8 = []  # sre
    with Neuron(VII, WMax, sos_fcon, sym, pre,
                id_=ntkh['common-6-7-8']):
        for i, somaloc in enumerate((L1, L2)):
            soma_index = i + 2  # L1 aligns to the 3rd the sypathetic ganglion
            # which is of course the L1 sypathetic ganglion, but it is
            # the 3rd ganlion in the model with a neuron 4 soma
            with Neuron(somaloc, *common[somaloc], id_=ntkh[f'common-6-7-8-{fn(somaloc)}']):
                with Neuron(n_ls, *common67[somaloc], id_=ntkh[f'common-6-7-{fn(somaloc)}-extra']):
                    # 6
                    n6 = kNeuron(IMGax, n_hg, PGsyn, ntk_6_ent, ntk_2_fcon,
                                 id_=ntku[f'6-{fn(somaloc)}'])
                    neuron_6.append(n6)
                    # 7
                    n7 = kNeuron(IMGsyn, ntk_7_ent, ntk_3_fcon,
                                 id_=ntku[f'7-{fn(somaloc)}'])
                    neuron_7.append(n7)

                # 8
                for syn_index, synloc in enumerate(four_soma_locs):
                    axons_in = syn_chain_axons_in(syn_index, soma_index)
                    syn = Phenotype(synloc, synp)
                    n8 = kNeuron(syn, *axons_in, ntk_8_ent, ntk_4_fcon,
                                 id_=ntku[f'8-{fn(somaloc)}-{fn(syn)}'])
                    neuron_8.append(n8)

    neuron_9 = []  # slm
    with Neuron(IX, WMax, n_pu, URTsyn, motor, id_=ntkh['common-9']):
        for somaloc, ventral_root_exit in zip((L5,    L6),
                                              (L5_vr, L6_vr)):
            n9 = kNeuron(somaloc, ventral_root_exit, ntk_9_ent, id_=ntku[f'9-{fn(somaloc)}'])
            neuron_9.append(n9)

    neuron_10 = []  # fos
    neuron_11 = []  # fos
    neuron_12 = []  # fos
    with Neuron(Isyn, IIsyn, Vsyn, VIIsyn, Xsyn, sens,
                id_=ntkh['common-10-11-12']):  # 10 11 12 XXX layers are wrong, intersection has to come before phenotype
        for somaloc in (L6_dr, S1_dr):
            n10 = kNeuron(somaloc, n_ps_dn, PGdn, n_bl_dn, ntk_10_ent,
                          id_=ntku[f'10-{fn(somaloc)}'])  # TODO a bunch of other sensory terminals
            neuron_10.append(n10)

            n12 = kNeuron(somaloc, n_pu_dn, URTdn, ntk_12_ent,
                          id_=ntku[f'12-{fn(somaloc)}'])  # TODO URTdn layer is rhabdosphincter
            neuron_12.append(n12)

        for somaloc in (L1_dr, L2_dr):
            n11 = kNeuron(somaloc, n_ls_dn, IMGdn, n_hg_dn, PGdn, n_bl_dn, ntk_11_ent,
                          id_=ntku[f'11-{fn(somaloc)}'])  # TODO bladder and layers
            neuron_11.append(n11)

    # thankfully alphabetical from head to toe
    with Neuron(pons_ax, id_=ntkh['axon-pons']):
        with Neuron(midbrain_ax, id_=ntkh['axon-midbrain']):
            with Neuron(dienc_ax, id_=ntkh['axon-dienceph']):
                with Neuron(cn_ax, id_=ntkh['axon-cernuc']):
                    neuron_13 = [kNeuron(BNST, BRGTNsyn, id_=ntkb['13'], **kld(13))]  # soma in BNST   ???
                    neuron_14 = [kNeuron(CeA, BRGTNsyn, id_=ntkb['14'], **kld(14))]   # soma in CeA    ???
                neuron_15 = [kNeuron(MPOA, BRGTNsyn, id_=ntkb['15'], **kld(15))]  # soma in MPOA   ???
                neuron_16 = [kNeuron(MnPO, BRGTNsyn, id_=ntkb['16'], **kld(16))]  # soma in MnPO   ???
                neuron_17 = [kNeuron(LPOA, BRGTNsyn, id_=ntkb['17'], **kld(17))]  # soma in LPOA   ???
                neuron_18 = [kNeuron(LHA, BRGTNsyn, id_=ntkb['18'], **kld(18))]   # soma in LHA    ???
            neuron_19 = [kNeuron(VLPAG, BRGTNsyn, id_=ntkb['19'], **kld(19))] # soma in VLPAG  ???
        neuron_20 = [kNeuron(BRGTN, id_=ntkb['20'], **kld(20))]  # soma in BRGTN  ??? # TODO many a syntapse


    # implicit types (parents) # FIXME XXX required modelling in many cases
    fos = first_order_sensory = [neuron_10, neuron_11, neuron_12]
    slm = somatic_lower_motor = [neuron_9]

    _pre = parasympathetic_pre_ganglionic = [neuron_5]
    pos = parasympathetic_post_ganglionic = [neuron_1]
    # para pre -> has soma located in some not sympathetic chain and projects to some ganglion that is also projected to by some sym pre
    # para post -> has reverse connection phenotype some para pre

    sre = sympathetic_pre_ganglionic = [neuron_6, neuron_7, neuron_8]
    sos = sympathetic_post_ganglionic = [neuron_2, neuron_3, neuron_4]
    # sym pre -> has forward connection phenotype some neuron population with cell body ni synaptic chain or other sympathetic ganglion???
    # sym post -> has reverse connection phenotype some sym pre
    # XXX these definitions are not good an collide with para I think

    #[print(repr(n)) for n in Neuron.neurons()]


ntks = {k: v
        for i in range(1,13)
        for (k, v) in (
                (f'ntk_{i}', Phenotype(npkb[str(i)], 'ilxtr:hasPhenotype')),
                (f'ntk_{i}_ent', EntailedPhenotype(npkb[str(i)], 'ilxtr:hasPhenotype')))}


class Keast2020(LocalNameManager):

    locals().update(ntks)  # LOL ?! this works !?

    # soma layers
    VII= Phenotype('UBERON:0016578', sllp)  # 'ilxtr:spinal-VII'
    IX = Phenotype('UBERON:0016580', sllp)  # 'ilxtr:spinal-IX'

    # soma locations

    #L1 = Phenotype('ilxtr:spinal-L1', slp)
    #L2 = Phenotype('ilxtr:spinal-L2', slp)
    #L5 = Phenotype('ilxtr:spinal-L5', slp)
    #L6 = Phenotype('ilxtr:spinal-L6', slp)
    #S1 = Phenotype('ilxtr:spinal-S1', slp)
    L1 = Phenotype('UBERON:0006448', slp)  # XXX fma in sct
    L2 = Phenotype('UBERON:0006450', slp)  # XXX fma in sct
    L5 = Phenotype('UBERON:0006447', slp)  # XXX fma in sct
    L6 = Phenotype('ILX:0738432', slp)  # 'ILX:0793358' was dupe
    S1 = Phenotype('UBERON:0006460', slp)  # XXX fma in sct

    IMG = Phenotype('UBERON:0005453', slp)  # ilxtr:IMG
    PG = Phenotype('UBERON:0016508', slp)  # ilxtr:PG

    #L1_dr = Phenotype('ilxtr:dr-L1', slp)  # soma implies ax + dn
    #L2_dr = Phenotype('ilxtr:dr-L2', slp)
    #L6_dr = Phenotype('ilxtr:dr-L6', slp)
    #S1_dr = Phenotype('ilxtr:dr-S1', slp)
    L1_dr = Phenotype('UBERON:0002855', slp)  # soma implies ax + dn
    L2_dr = Phenotype('UBERON:0002856', slp)
    L6_dr = Phenotype('ILX:0793360', slp)
    S1_dr = Phenotype('UBERON:0002860', slp)

    # sort of nerves
    L1_vr = Phenotype('ILX:0785421', alp)  # ventral root 'ilxtr:vr-L1'  # XXX fma in sct
    L1_wr = Phenotype('ILX:0793220', alp)  # white ramus 'ilxtr:wr-L1'  # XXX FMA in sct

    L1_gr = Phenotype('ILX:0785825', alp)  # gray ramus 'ilxtr:gr-L1'  # XXX fma in sct

    L2_vr = Phenotype('ILX:0788675', alp)  # 'ilxtr:vr-L2'  # XXX fma in sct
    L2_wr = Phenotype('ILX:0793221', alp)  # 'ilxtr:wr-L2'  # XXX FMA in sct

    L2_gr = Phenotype('ILX:0785733', alp)  # 'ilxtr:gr-L2'  # XXX fma in sct

    L5_vr = Phenotype('ILX:0791148', alp)  # 'ilxtr:vr-L5'  # XXX fma in sct

    L6_gr = Phenotype('ILX:0739299', alp)  # 'ilxtr:gr-L6'
    L6_vr = Phenotype('ILX:0793615', alp)  # 'ilxtr:vr-L6'

    S1_gr = Phenotype('ILX:0793228', alp)  # 'ilxtr:gr-S1'  # XXX fma in sct
    S1_vr = Phenotype('ILX:0792853', alp)  # 'ilxtr:vr-S1'  # XXX fma in sct

    IMGax = Phenotype('UBERON:0005453', alp)
    PGax = Phenotype('UBERON:0016508', alp)

    S1ax = Phenotype('ILX:0793350', alp)  # 'ilxtr:sc-S1'

    # ilxtr:spinal-white-matter technically a layer
    # FIXME because axon location does not have cardianlity 1 WMax is
    # ambiguous unless it is composed with a layer of the spinal cord
    # because the operation is not commutative with card > 1
    #WMax = Phenotype('ilxtr:spinal-white-matter', alp)
    WMax = Phenotype('UBERON:0002318', alp)


    # sensory dendrite sort of nerves
    IMGdn = Phenotype('UBERON:0005453', dlp)
    PGdn = Phenotype('UBERON:0016508', dlp)

    # XXX FIXME I'm being something of a literalist here, and guessing that
    # the boxes that are drawn for these might not actually be what is intended
    cn_ax = Phenotype('UBERON:0010011', alp)  # collection of basal ganglia, syn cerebral nuclei ??
    dienc_ax = Phenotype('UBERON:0001894', alp)
    midbrain_ax = Phenotype('UBERON:0001891', alp)
    pons_ax = Phenotype('UBERON:0000988', alp)
    medulla_ax = Phenotype('UBERON:0001896', alp)

    # nerves
    #n_bl = Phenotype('ilxtr:nerve-bladder', alp)
    #n_hg = Phenotype('ilxtr:nerve-hypogastric', alp)
    #n_ls = Phenotype('ilxtr:nerve-lumbar-splanic', alp)
    #n_ps = Phenotype('ilxtr:nerve-pelvic-splanic', alp) # splanchnic
    #n_pu = Phenotype('ilxtr:nerve-pudendal', alp)
    n_bl = Phenotype('ILX:0793559', alp)
    n_hg = Phenotype('UBERON:0005303', alp)
    n_ls = Phenotype('UBERON:0018683', alp)
    n_ps = Phenotype('UBERON:0018675', alp)
    n_pu = Phenotype('UBERON:0011390', alp)

    # sensory ??ents
    #n_bl_dn = Phenotype('ilxtr:nerve-bladder', dlp)
    #n_hg_dn = Phenotype('ilxtr:nerve-hypogastric', dlp)
    #n_ls_dn = Phenotype('ilxtr:nerve-lumbar-splanic', dlp)
    #n_ps_dn = Phenotype('ilxtr:nerve-pelvic-splanic', dlp)
    #n_pu_dn = Phenotype('ilxtr:nerve-pudendal', dlp)
    n_bl_dn = Phenotype('ILX:0793559', dlp)
    n_hg_dn = Phenotype('UBERON:0005303', dlp)
    n_ls_dn = Phenotype('UBERON:0018683', dlp)
    n_ps_dn = Phenotype('UBERON:0018675', dlp)
    n_pu_dn = Phenotype('UBERON:0011390', dlp)

    #URTdn = Phenotype('ilxtr:urethra', dlp)
    URTdn = Phenotype('UBERON:0000057', dlp)

    # synaptic locations
    IMGsyn = Phenotype('UBERON:0005453', synp)
    PGsyn = Phenotype('UBERON:0016508', synp)

    BNVWsyn = Phenotype('ilxtr:bladder-neck-vessel-wall', synp)
    BDVWsyn = Phenotype('ilxtr:bladder-dome-vessel-wall', synp)

    BDWsyn = Phenotype('ilxtr:bladder-dome-wall', synp)
    BNWsyn = Phenotype('ILX:0774144', synp)  # 'ilxtr:bladder-neck-wall'

    #URTsyn = Phenotype('ilxtr:urethra', synp)
    URTsyn = Phenotype('UBERON:0000057', synp)

    BNST = Phenotype('UBERON:0001880', slp)
    CeA = Phenotype('UBERON:0002883', slp)  # central amygdala
    MPOA = Phenotype('UBERON:0007769', slp)
    MnPO = Phenotype('UBERON:0002625', slp)
    LPOA = Phenotype('UBERON:0001931', slp)
    LHA = Phenotype('UBERON:0002430', slp)
    VLPAG = Phenotype('ILX:0793626', slp) # vent lat peri aq gray
    BRGTN = Phenotype('UBERON:0007632', slp) # barrington's nucleus

    BRGTNsyn = Phenotype('UBERON:0007632', synp)

    # FIXME synp into layer has the cardinality issue
    # with the fact that (intersection (phenotype layer hp) (phenotype region hp))
    # is not commutative with (phenotype (intersection layer region) hp)
    # when the relationship hp allows cardinality > 1

    #Isyn = Phenotype('ilxtr:spinal-I', synp)
    #IIsyn = Phenotype('ilxtr:spinal-II', synp)
    #Vsyn = Phenotype('ilxtr:spinal-V', synp)
    #VIIsyn = Phenotype('ilxtr:spinal-VII', synp)
    #Xsyn = Phenotype('ilxtr:spinal-X', synp)
    Isyn = Phenotype('UBERON:0006118', synp)
    IIsyn = Phenotype('ILX:0110092', synp)
    Vsyn = Phenotype('UBERON:0016576', synp)
    VIIsyn = Phenotype('UBERON:0016578', synp)
    Xsyn = Phenotype('ILX:0110100', synp)

    # sensory substructure locations

    # target cell types
    #ntk_4 = Phenotype('ilxtr:neuron-phenotype-keast-4', 'ilxtr:hasPhenotype')  # FIXME what is this really?
    #ntk_4_ent = EntailedPhenotype('ilxtr:neuron-phenotype-keast-4', 'ilxtr:hasPhenotype')  # FIXME what is this really?
    # XXX using a specific phenotype here is undesireable, because it must _always_ be asserted
    # in order for membership to be determined, this is true for individuals as well, therefore this
    # should be used as entailed restriction so that anything that matches the necessary contitions
    # will automatically classify as a result

    # is this a phenotype or a superclass? XXX kind of dirty to do it this way by creating an aribrary
    # phenotype that is borne by these, but being marked as a keast-type-4 neuron is a thing so here we are
    # in a pure type world, this would be the coloring and the numbering, but it is more assertional
    # making it a superclass is probably better than

    ntk_1_fcon = Phenotype('ilxtr:neuron-type-keast-1', fconp)
    ntk_2_fcon = Phenotype('ilxtr:neuron-type-keast-2', fconp)
    ntk_3_fcon = Phenotype('ilxtr:neuron-type-keast-3', fconp)
    ntk_4_fcon = Phenotype('ilxtr:neuron-type-keast-4', fconp)

    # FIXME this is super confusing, but this id is equivalent to all
    # neurons that bear the ilxtr:neuron-phenotype-sym-post phenotype
    sos_fcon = Phenotype('ilxtr:sympathetic-post-ganglionic', fconp)  # FIXME likely overly broad?

    # (para)?sympathetic colorings (until we can get deeper modelling correct)
    #para_pre = Phenotype('ilxtr:neuron-phenotype-para-pre')
    #para_post = Phenotype('ilxtr:neuron-phenotype-para-post')
    #sym_pre = Phenotype('ilxtr:neuron-phenotype-sym-pre')
    #sym_post = Phenotype('ilxtr:neuron-phenotype-sym-post')

    para = Phenotype('ilxtr:ParasympatheticPhenotype')
    sym = Phenotype('ilxtr:SympatheticPhenotype')
    pre = Phenotype('ilxtr:PreGanglionicPhenotype')
    post = Phenotype('ilxtr:PostGanglionicPhenotype')

    sens = Phenotype('ilxtr:SensoryPhenotype')
    motor = Phenotype('ilxtr:MotorPhenotype')
    #intrin = Phenotype('ilxtr:IntrinsicPhenotype')



    # phenotypes over layers do not compose well because the
    # intersection of the layer is ambiguous with regard to which
    # other anatomical structures it applies to, so for example and it
    # does not sufficiently constrain that it must be the same axon
    # segment that is in both the bladder dome wall and epithelium we
    # do not currently have a way to express anatomical intersections
    # in neuron lang, we were able to punt on this for somas because
    # there is at most 1 soma, so the referent is unambiguous

    #SMsyn = Phenotype('ilxtr:smooth-muscle', synp)
    #EPIsyn = Phenotype('ilxtr:epithelium', synp)
    #RABsyn = Phenotype('ilxtr:epithelium', synp)

four_soma_locs = [
    # sympathetic chain
    'ILX:0787009',  # 'ilxtr:sc-T12'
    'ILX:0739295',  # 'ilxtr:sc-T13'
    'ILX:0789862',  # 'ilxtr:sc-L1'
    'ILX:0786933',  # 'ilxtr:sc-L2'
    'ILX:0788315',  # 'ilxtr:sc-L3'
    'ILX:0786049',  # 'ilxtr:sc-L4'
    'ILX:0739296',  # 'ilxtr:sc-L5'
    'ILX:0790472',  # 'ilxtr:sc-L6'
    #'ILX:0789109',  # 'ilxtr:sc-S1'
]

osl = {
    'ILX:0787009': 'ilxtr:sc-T12',
    'ILX:0739295': 'ilxtr:sc-T13',
    'ILX:0789862': 'ilxtr:sc-L1',
    'ILX:0786933': 'ilxtr:sc-L2',
    'ILX:0788315': 'ilxtr:sc-L3',
    'ILX:0786049': 'ilxtr:sc-L4',
    'ILX:0739296': 'ilxtr:sc-L5',
    'ILX:0790472': 'ilxtr:sc-L6',
    #'ILX:0789109': 'ilxtr:sc-S1',
}

_osl_chain = {
    'ILX:0793342': 'ilxtr:sc-T12',
    'ILX:0793343': 'ilxtr:sc-T13',
    'ILX:0793344': 'ilxtr:sc-L1',
    'ILX:0793345': 'ilxtr:sc-L2',
    'ILX:0793346': 'ilxtr:sc-L3',
    'ILX:0793347': 'ilxtr:sc-L4',
    'ILX:0793348': 'ilxtr:sc-L5',
    'ILX:0793349': 'ilxtr:sc-L6',
    # 'ILX:0793350': 'ilxtr:sc-S1',
}

[setattr(Keast2020, f'sc{osl[sl].split("-")[-1]}',
         Phenotype(sl, slp))
 for sl in four_soma_locs]

[setattr(Keast2020, f'sc{osl[sl].split("-")[-1]}ax',
         Phenotype(sl, alp))
 for sl in four_soma_locs]

[setattr(Keast2020, f'sc{osl[sl].split("-")[-1]}syn',
         Phenotype(sl, synp))
 for sl in four_soma_locs]

sympathetic_chain_axons = [Phenotype(l, alp) for l in four_soma_locs]  # FIXME intersegmental


def syn_chain_axons_in(syn_index, soma_index):
    """ determine which regions a single collateral must pass through """
    if syn_index < soma_index:
        axons_in = sympathetic_chain_axons[syn_index + 1:soma_index + 1]
    elif syn_index == soma_index:
        axons_in = sympathetic_chain_axons[soma_index],
    else:
        axons_in = sympathetic_chain_axons[soma_index:syn_index]

    return axons_in


def main():
    from pyontutils.utils import relative_path
    config = Config('keast-2020',
                    source_file=relative_path(__file__, no_wd_value=__file__))

    with Keast2020:
        needs_keast_namespace(config)

    config.write()
    labels = (
        rdfs.label,
        #ilxtr.genLabel, ilxtr.localLabel, ilxtr.simpleLabel,
        #ilxtr.simpleLocalLabel, skos.prefLabel
    )
    to_remove = [t for t in config._written_graph
                 if t[1] in labels and '/neuron-type-keast-' in t[0]]
    [config._written_graph.remove(t) for t in to_remove]
    config._written_graph.write()
    config.write_python()
    return config,


if __name__ == '__main__':
    main()
else:
    __globals__ = globals()



# pytest module discovery somehow adds the nested compiled module
# and models module to sys.modules which breaks tests, so we run
# this file as the first test after test collection
# I'm sure there is a non-hacked sanctioned way to actually do this
# but for now let's not worry about that
import sys


def test_fix_sys_modules():
    sys.modules.pop('compiled', None)
    sys.modules.pop('models', None)
